import codecs
import struct
import re
from PyQt5.QtGui import QFont, QFontDatabase
from PyQt5.QtWidgets import QProgressBar, QApplication
from copy import deepcopy
from datetime import datetime
from pyparsing import (
    Literal, 
    Word, 
    printables, 
    alphas, 
    nums, 
    hexnums,
    common,
    White,
    replace_with,
    Opt, 
    Or,
    one_of,
    Combine, 
    OneOrMore, 
    Group,
    Suppress,
    Forward,
    ParseResults
)

from plover import log

from plover_cat.steno_objects import *
# control chars \ { }


LBRACE, RBRACE, BKS = map(Suppress, "{}\\")


text_whitespace = Opt(White(" \t")) + Word(printables, exclude_chars="\\{}") + Opt(White(" \t"))
text_whitespace.set_name("text")
# text_whitespace.leave_whitespace()

# special chars that exist as control words, should be replaced by actual unicode equivalents, 
tab_char = Literal("\\tab").set_parse_action(replace_with("\N{CHARACTER TABULATION}")) # "\t" or "\N{CHARACTER TABULATION}"
emdash_char = Literal("\\emdash").set_parse_action(replace_with("\N{EM DASH}")) # "\u2014" or "\N{EM DASH}"
endash_char = Literal("\\endash").set_parse_action(replace_with("\N{EN DASH}")) # "\u2013" or "\N{EN DASH}"
emspace_char = Literal("\\emspace").set_parse_action(replace_with("\N{EM SPACE}")) # "\u2003" or "\N{EM SPACE}"
enspace_char = Literal("\\enspace").set_parse_action(replace_with("\N{EN SPACE}")) # "\u2002" or "\N{EN SPACE}"
nonbreaking_space = Literal("\\~").set_parse_action(replace_with("\N{NO-BREAK SPACE}")) # "\u00A0" or "\N{NO-BREAK SPACE}"
soft_hyphen = Literal("\\-").set_parse_action(replace_with("\N{SOFT HYPHEN}")) # "\u00AD" or "\N{SOFT HYPHEN}"
nonbreaking_hyphen = Literal("\\_").set_parse_action(replace_with("\N{HYPHEN}")) # "\u2010" or "\N{HYPHEN}"
bullet_char = Literal("\\bullet").set_parse_action(replace_with("\N{BULLET}"))
lquote_char = Literal("\\lquote").set_parse_action(replace_with("\N{LEFT SINGLE QUOTATION MARK}"))
rquote_char = Literal("\\rquote").set_parse_action(replace_with("\N{RIGHT SINGLE QUOTATION MARK}"))
ldblquote_char = Literal("\\ldblquote").set_parse_action(replace_with("\N{LEFT DOUBLE QUOTATION MARK}"))
rdblquote_char = Literal("\\rdblquote").set_parse_action(replace_with("\N{RIGHT DOUBLE QUOTATION MARK}"))
open_brace = Literal("\\{")
close_brace = Literal("\\}")

def pa(s, l, t):
    hex_val = [int(tokn, 16) for tokn in t]
    byte_val = struct.pack('B', int(hex_val[0]))
    return(byte_val.decode("cp437"))

hex_char = Group(Suppress(BKS + "'") + Word(hexnums).set_parse_action(pa).set_results_name("hex_char")) # hex in form \'xx
hex_char.set_name("Hexadecimal character")

def paste_text(s, l, t):
    command_dict = {}
    command_dict["control"] = "text"
    command_dict["value"] = "".join(t[0])
    return(command_dict)

text_and_chars = Group(OneOrMore(text_whitespace ^ tab_char ^ emdash_char ^ endash_char ^ 
    emspace_char ^ enspace_char ^ nonbreaking_space ^ soft_hyphen ^ 
    nonbreaking_hyphen ^ hex_char ^ open_brace ^ close_brace ^ bullet_char ^ 
    lquote_char ^ rquote_char ^ ldblquote_char ^ rdblquote_char))
text_and_chars.set_name("text")
text_and_chars.set_name("Plain text")
text_and_chars.set_parse_action(paste_text)

expr = Forward()

def control_parse(s, l, t):
    command_dict = {}
    command_dict["control"] = t[0]["control"]
    try:
        command_dict["value"] = t[0]["num"]
    except KeyError:
        pass
    try:
        t[0]["ending"]
        command_dict["ending"] = True
    except KeyError:
        pass
    return(command_dict)

control = Group(
    BKS + 
    Combine(
        Word(alphas)("control") + 
        Opt(Word(nums + '-').set_parse_action(common.convert_to_integer))("num") + 
        Opt(White(" ", max=1)) + 
        Opt(Literal(";"))("ending")
        )
    )
control.set_name("control")
control.set_parse_action(control_parse)

def control_parse_ignore(s, l, t):
    command_dict = t[0][1]
    command_dict["ignore"] = True
    return(command_dict)

ignore = Group(Literal("\\*")("ignore") + control)
ignore.set_name("ignore")
ignore.set_parse_action(control_parse_ignore)

expr <<= OneOrMore(ignore("ignore*") | hex_char("hex_char*") | control("control*") | text_and_chars("text*") | Group(LBRACE + expr + RBRACE)("group*"))
expr.set_name("RTF")

style_string = Suppress(Literal("\\par\\pard")) + OneOrMore(control("control*"))

def twip_to_in(twips):
    return(twips / 1440)

def in_to_twip(inches):
    if isinstance(inches, str):
        inches = float(inches.replace("in", ""))
    return(int(inches * 1440))

def append_value(dict_obj, key, value):
    # Check if key exist in dict or not
    if key in dict_obj:
        # Key exist in dict.
        # Check if type of value of key is list or not
        if not isinstance(dict_obj[key], list):
            # If type is not list then make it list
            dict_obj[key] = [dict_obj[key]]
        # Append the value in list
        dict_obj[key].append(value)
    else:
        # As key is not in dict,
        # so, add key-value pair
        dict_obj[key] = value

def collapse_dict(element):
    new_dict = {}
    for i in element:
        if "value" in i:
            append_value(new_dict, i["control"], i["value"])
        else:
            append_value(new_dict, i["control"], "")
    return(new_dict)

# with rtf_steno parsing, each control word parses as a dict, and text as a dict
# each group parses as a list (ParseResults), and so results in a list with control words as dicts, more groups as lists, and text as dicts
class rtf_steno:
    def __init__(self, file_name, progress_bar = None):
        self.rtf_file = file_name
        self.parse_results = None
        self.scanned_styles = {}
        self.framerate = 30
        self.fonts = {}
        self.styles = {}
        self.page = {}
        # mimic json_document format in this dict
        self.paragraphs = {}
        # list of strokes, within par, empty at next par and append to paragraphs 
        self.par = []
        self.par_style = ""
        self.date = ""
        # dict of timecodes to bump off
        self.timecode = { "milli": "000", "sec": "00", "min": "00", "hour": "00"}
        self.steno = ""
        self.text = ""
        self.start_parsing_text = False
        self.defaultfont = ""
        self.fonts = {}
        self.progress_bar = progress_bar
    def parse_framerate(self, element):
        element_dict = element[0]
        self.framerate = element_dict["value"]
    def parse_date(self, element):
        for i in element:
            try:
                control_word = i[0]["control"]
                if control_word == "creatim":
                    select_dict = i
                    break
            except:
                pass
        for i in select_dict:
            if i["control"] == "yr":
                year = i["value"]
            elif i["control"] == "mo":
                month = i["value"]
            elif i["control"] == "dy":
                day = i["value"]
        self.date = "%s-%s-%s" % (year, month, day)
    def parse_font(self, element):
        for i in element[1:len(element)]:
            new_font_dict = collapse_dict(i)
            self.fonts[str(new_font_dict["f"])] = new_font_dict
    def parse_styles(self, element):
        for i in element[1:len(element)]:
            new_style_dict = collapse_dict(i)
            # libreoffice replaces spaces with _20_ in style name
            split_style_name = new_style_dict["text"]
            correct_style_name = "_20_".join(split_style_name.split(" ")).replace(";", "")
            new_style_dict["text"] = correct_style_name
            self.styles[str(new_style_dict["s"])] = new_style_dict
    def parse_page(self, element):
        self.page[element["control"]] = round(twip_to_in(element["value"]), 2)
    def parse_timecode(self, element, split = ":"):
        time_element = element[1]["value"].split(split)
        # previous timecode is structured to take advantage of dict being in order
        # plover is python 3.7 and above so this should work properly
        split_codes = ["%02d" % int(i) for i in reversed(time_element)]
        # remember that last element in original timecode is frame, needs millisecond
        split_codes[0] = "%03d" % self.convert_framerate_milli(int(split_codes[0]))
        for key_index, text in enumerate(split_codes):
            self.timecode[list(self.timecode)[key_index]] = text
    def parse_par_style(self, element):
        style_index = str(element["value"])
        style_name = self.styles[style_index]["text"]
        self.par_style = style_name
    def parse_cxa(self, element):
        cxa_dict = collapse_dict(element)
        self.steno = ""
        self.text = cxa_dict["text"]
        timestamp = "%sT%s:%s:%s.%s" % (self.date, self.timecode["hour"], self.timecode["min"], self.timecode["sec"], self.timecode["milli"])
        ael = automatic_text(prefix = cxa_dict["text"], time = timestamp)
        self.par.append(ael.to_json())
    def parse_steno(self, element):
        try:
            stroke = element[1]["value"]
        except:
            stroke = ""
        self.steno = stroke
    def parse_text(self, element):
        self.text = element["value"]
    def append_stroke(self):
        timestamp = "%sT%s:%s:%s.%s" % (self.date, self.timecode["hour"], self.timecode["min"], self.timecode["sec"], self.timecode["milli"])
        ael = stroke_text(time = timestamp, stroke = self.steno, text = self.text)
        self.par.append(ael.to_json())
    def convert_framerate_milli(self, frames):
        milli = 1000 * frames / self.framerate
        return(milli)
    def set_new_paragraph(self):
        par_num = str(len(self.paragraphs))
        # par_text = "".join([stroke[2] for stroke in self.par])
        # this last stroke should capture the stroke emitting \par
        timestamp = "%sT%s:%s:%s.%s" % (self.date, self.timecode["hour"], self.timecode["min"], self.timecode["sec"], self.timecode["milli"])
        last_stroke = stroke_text(time = timestamp, stroke = self.steno, text = "\n")
        self.par.append(last_stroke.to_json())
        strokes = self.par
        par_dict = {}
        par_dict["strokes"] = strokes
        par_dict["creationtime"] = strokes[0]["time"]
        par_dict["style"] = self.par_style
        self.paragraphs[par_num] = par_dict
        self.par = []
    def parse_document(self):
        parse_results = expr.parse_file(self.rtf_file)
        if self.progress_bar:
            self.progress_bar.setMaximum(len(parse_results[0]))
        self.parse_results = parse_results
        for num, i in enumerate(parse_results[0]):
            if isinstance(i, dict):
                command_name = i["control"]
                if command_name in ["paperh", "paperw", "margt", "margb", "margl", "margr"]:
                    self.parse_page(i)
                if command_name == "deffont":
                    self.defaultfont = i["value"] # this is a number representing index into fonts
                if command_name == "par":
                    if not self.start_parsing_text:
                        self.start_parsing_text = True
                    else:
                        self.set_new_paragraph()
                if command_name == "s":
                    self.parse_par_style(i)
                if command_name == "text" and self.start_parsing_text:
                    self.parse_text(i)
                    self.append_stroke()
            elif isinstance(i, ParseResults):
                command_name = i[0]["control"]
                if command_name == "stylesheet":
                    self.parse_styles(i)
                if command_name == "fonttbl":
                    self.parse_font(i)
                if command_name == "info":
                    self.parse_date(i)
                if command_name == "cxframes":
                    self.parse_framerate(i)
                if command_name == "cxt":
                    self.parse_timecode(i)
                if command_name == "cxs":
                    self.parse_steno(i)
                if command_name == "cxa":
                    self.parse_cxa(i)
            else:
                pass
            if self.progress_bar:
                self.progress_bar.setValue(num)
                QApplication.processEvents()
        self.set_new_paragraph()
        self.scan_par_styles()
    def scan_par_styles(self):
        with open(self.rtf_file, 'r') as f:
            data = f.read().rstrip()
        style_list = []
        for i in style_string.scanString(data):
            style_list.append(i[0].asList())
        # print(len(style_list))
        par_style_index = 0
        for ind, el in enumerate(style_list):
            new_style_dict = collapse_dict(el)
            # print(new_style_dict)
            if "s" in new_style_dict:
                self.scanned_styles[str(par_style_index)] = new_style_dict
                par_style_index += 1
                
# test_string = """
# {\\rtf1\\ansi\\deff1
# {\\fonttbl
# {\\f0\\fcharset1 Times New Roman;}
# {\\f1\\froman\\fcharset1 Times New Roman{\\*\\falt Base Font};}}
# {\\colortbl;\\red0\\green0\\blue0;\\red0\\green0\\blue255;\\red0\\green255\\blue255;\\red0\\green255\\blue0;}
# {\\xe{\\*\cxexnum 1}Exhibit 1:  A knife }} sss
# """
# test_string = r"{\par\pard\s0\f0\fs24{\*\cxt 20:46:25:00}{\*\cxs WELG}Welcome{\*\cxt 20:46:26:00}{\*\cxs TOT} to the{\*\cxt 20:46:27:00}{\*\cxs SROEUS} Voice{\*\cxt 20:46:27:00}{\*\cxs -F} of{\*\cxt 20:46:27:00}{\*\cxs PHERBG} America{\xe{\*\cxexnum 1}Exhibit 1:  A knife }{\*\cxt 20:46:28:00}{\*\cxs AES}'s{\*\cxt 20:46:29:00}{\*\cxs TPHUS} News{\*\cxt 20:46:30:00}{\*\cxs WORDZ} Words{\*\cxt 20:46:30:00}{\*\cxs TP-PL}.}"


# res = expr.parse_string(test_string)
# test_rtf = rtf_steno("plover_cat/test.rtf")
# test_rtf.parse_document()
# test_rtf.scan_par_styles()
# test_rtf.parse_results[0][9]

def rtf_to_qfont(font):
    font_dict = {}
    if "text" in font:
        qt_font = QFont(text)
    else:
        qt_font = QFont()
    if "froman" in font:
        qt_font.setStyleHint(QFont.Times)
    if "fswiss" in font:
        qt_font.setStyleHint(QFont.SansSerif)
    if "fmodern" in font:
        qt_font.setStyleHint(QFont.Courier)
    if "fscript" in font:
        qt_font.setStyleHint(QFont.Cursive)
    if "fdecor" in font:
        qt_font.setStyleHint(QFont.Decorative)
    return(qt_font)

def modify_styleindex_to_name(styles, style_names):
    for key, i in styles.items():
        if "parentstylename" in i:
            style_names[i["parentstylename"]]
            i["parentstylename"] = style_names[i["parentstylename"]]
        if "nextstylename" in i:
            i["nextstylename"] = style_names[int(i["nextstylename"])]
        styles[key] = i
    return(styles)

def extract_par_style(style_dict):
    one_style_dict = {}
    one_part_dict = {}
    one_text_dict = {}
    if "s" in style_dict:
        one_style_dict["styleindex"] = str(style_dict["s"])
    if "sbasedon" in style_dict:
        one_style_dict["parentstylename"] = style_dict["sbasedon"]
    if "snext" in style_dict:
        one_style_dict["nextstylename"] = style_dict["snext"]
    if "li" in style_dict:
        one_part_dict["marginleft"] = "%.2fin" % twip_to_in(style_dict["li"])
    if "ri" in style_dict:
        one_part_dict["marginright"] = "%.2fin" % twip_to_in(style_dict["ri"])
    if "fi" in style_dict:
        one_part_dict["textindent"] = "%.2fin" % twip_to_in(style_dict["fi"])
    if "ql" in style_dict:
        one_part_dict["textalign"] = "left"
    if "qr" in style_dict:
        one_part_dict["textalign"] = "right"
    if "qj" in style_dict:
        one_part_dict["textalign"] = "justify"
    if "qc" in style_dict:
        one_part_dict["textalign"] = "center"
    if "sb" in style_dict:
        one_part_dict["margintop"] = "%.2fin" % twip_to_in(style_dict["sb"])
    if "sa" in style_dict:
        one_part_dict["marginbottom"] = "%.2fin" % twip_to_in(style_dict["sa"])
    if "tx" in style_dict:
        if isinstance(style_dict["tx"], list):
            tab_pos = ["%.2fin" % twip_to_in(i) for i in style_dict["tx"]]
        else:
            tab_pos = ["%.2fin" % twip_to_in(style_dict["tx"])]
        one_part_dict["tabstop"] = tab_pos
    if "f" in style_dict:
        one_text_dict["fontindex"] = style_dict["f"]
    if "fs" in style_dict:
        one_text_dict["fontsize"] = str(style_dict["fs"]/2)
    if "i" in style_dict:
        one_text_dict["fontstyle"] = "italic"
    if "b" in style_dict:
        one_text_dict["fontweight"] = "bold"
    if "ul" in style_dict:
        one_text_dict["textunderlinetype"] = "single"
        one_text_dict["textunderlinestyle"] = "solid"
    one_style_dict["paragraphproperties"] = one_part_dict
    one_style_dict["textproperties"] = one_text_dict
    return(one_style_dict)

def load_rtf_styles(parse_results):
    log.debug(f"Loading RTF styles from {parse_results.rtf_file}")
    style_dict = {}
    for k, v in parse_results.styles.items():
        style_name = v["text"]
        style_num = 0
        while style_name in list(style_dict.keys()):
            style_num += 1
            style_name = re.sub("\\d+$", "", style_name) + str(style_num)
        style_dict[style_name] = extract_par_style(v)
    styles = []
    for k, v in style_dict.items():
        styles.append(k)
    style_dict = modify_styleindex_to_name(style_dict, styles)
    font_names = []
    for i, font in parse_results.fonts.items():
        font_names.append(font["text"])
    indiv_style = {}
    for key, par_style in parse_results.scanned_styles.items():
        par_dict = extract_par_style(par_style)
        # indiv_style[key] = extract_par_style(par_style)
        if "textproperties" in par_dict and "fontindex" in par_dict["textproperties"]:
            fontindex = str(par_dict["textproperties"]["fontindex"])
            font_dict = parse_results.fonts[fontindex]
            if "text" in font_dict:
                par_dict["textproperties"]["fontname"] = font_dict["text"].replace(";", "")
                par_dict["textproperties"]["fontfamily"] = font_dict["text"].replace(";", "")
            if "froman" in font_dict:
                par_dict["textproperties"]["fontfamilygeneric"] = "roman"
            if "fswiss" in font_dict:
                par_dict["textproperties"]["fontfamilygeneric"] = "swiss"
            if "fmodern" in font_dict:
                par_dict["textproperties"]["fontfamilygeneric"] = "modern"
            if "fscript" in font_dict:
                par_dict["textproperties"]["fontfamilygeneric"] = "script"
            if "fdecor" in font_dict:
                par_dict["textproperties"]["fontfamilygeneric"] = "decorative"
        indiv_style[key] = par_dict
    renamed_indiv_style = []
    for index, style in indiv_style.items():
        par_style = style
        doc_style = deepcopy(list(style_dict.values())[int(style["styleindex"])])
        doc_style.update(par_style)
        found_style = False
        for k, v in style_dict.items():
            if doc_style == v:
                renamed_indiv_style.append(k)
                found_style = True
                break
        if found_style:
            continue
        style_parent_name = list(style_dict.keys())[int(style["styleindex"])]
        detect_int = re.search("\\d+$", style_parent_name)
        if detect_int:
            style_num = int(detect_int.group()) + 1
        else:
            style_num = 0
        new_style_name = re.sub("\\d+$", "", style_parent_name) + str(style_num)
        while new_style_name in list(style_dict.keys()):
            style_num += 1
            new_style_name = re.sub("\\d+$", "",style_parent_name) + str(style_num)
        style_dict[new_style_name] = doc_style
        renamed_indiv_style.append(new_style_name)
    return(style_dict, renamed_indiv_style)    




from urllib import request, parse
from urllib.error import HTTPError
from PyQt5.QtCore import QObject, pyqtSignal
from queue import Queue, Empty
from collections import deque
from plover_cat.steno_objects import wordsep_simple_re
from datetime import datetime
import obsws_python as obs
from obsws_python.error import OBSSDKRequestError

class captionWorker(QObject):
    """Generate captions based on user settings and endpoint parameters 
    for caption display and to send to endpoints. 
    
    ``captionWorker`` is put into another thread and doesn't run on the main event thread.
    Text gets ingested and then sent out as formatted caption lines.

    :param max_length: maximum number of characters for each caption line, 
        suggested value 32, default None
    :type max_length: int, optional
    :max_lines: maximum number of captions lines to display, 
        suggested value 3, default None
    :type max_lines: int, optional
    :param remote: name of remote endpoint, can only be one of supported, default None
    :type remote: str, optional
    :param endpoint: URL or local port depending on what kind of remote, may be authentication token,
        suggested localhost in the case of OBS, default None
    :type endpoint: str, optional
    :param port: port number to use with endpoint, suggested 4455 for OBS, default None
    :type port: str, optional
    :param password: password to use along with other fields above, default None
    :type password: str, optional
    """    
    capSend = pyqtSignal(str)
    """Signal sent with formatted caption line."""
    finished = pyqtSignal()
    """Signal sent when worker is done."""
    postMessage = pyqtSignal(str)
    """Signal sent with message to display."""
    def __init__(self, max_length = None, max_lines = None, remote = None, endpoint = None, port = None, password = None):
        QObject.__init__(self)
        self.max_length = max_length
        self.max_lines = max_lines
        self.remote = remote
        self.endpoint = endpoint
        self.port = port
        self.password = password
        if self.remote == "OBS":
            if not self.port:
                self.port = "4455"
            if not self.endpoint:
                self.endpoint = "localhost"
            self.obs = obs.ReqClient(host=self.endpoint, port=self.port, password=self.password, timeout=3) 
            self.obs_queue = deque(maxlen = self.max_lines)
        self.word_queue = Queue()
        """Queue containing text split into word chunks."""
        self.cap_queue = deque(maxlen = max_lines)
        """Queue containing formatted caption lines ready for display."""
        self.next_word = "" 
        self.zoom_seq = 1
        """Line number for caption, required for Zoom captions."""
    def intake(self, text):
        """Receive text from main editor and create work chunks for captions
        :param text: text written into editor
        :type text: str, required
        """
        text_time = datetime.now()
        chunks = [i for i in wordsep_simple_re.split(text) if i]
        for i in chunks:
            # tuple, first text, then time
            self.word_queue.put((i, text_time))
        self.make_caps()
    def make_caps(self):
        """Ingest word chunks from queue and put lines into caption queue.
        """
        while not self.word_queue.empty():
            self.next_word, text_time = self.word_queue.get()
            try:
                last_cap, cap_time = self.cap_queue.pop()
            except IndexError:
                last_cap = ""
            if len(last_cap) + len(self.next_word) > self.max_length:
                self.cap_queue.append((last_cap, cap_time))
                self.cap_queue.append((self.next_word, text_time))
            elif "\u2029" in self.next_word:
                self.cap_queue.append((last_cap, cap_time))
                self.cap_queue.append((self.next_word.replace("\u2029", ""), text_time))
            else:
                self.cap_queue.append((last_cap + self.next_word, text_time))
            self.send_cap()
    def send_cap(self):
        """Take caption from queue and send to display, and also if endpoint is defined.
        """
        try:
            last_caps = list(self.cap_queue)
            cap = "\n".join(c[0].lstrip(" ") for c in last_caps if c[0].lstrip(" "))
            self.capSend.emit(cap)
            if self.endpoint:
                if self.remote == "Microsoft Teams":
                    self.send_msteams(cap)
                elif self.remote == "Zoom":
                    self.send_zoom(cap)
                elif self.remote == "OBS":
                    self.send_obs(cap)
        except Empty:
            pass
    def clean_and_stop(self):
        """Clean up instance and emit signal for finish
        """
        self.finished.emit()
    def send_msteams(self, cap):
        """Take caption and send post to Microsoft Teams session
        """
        req = request.Request(self.endpoint, method = "POST")
        req.data = cap.encode()
        req.add_header('Content-Type', 'text/plain')
        req.add_header('Content-Length', len(cap)) 
        try:
            r = request.urlopen(req)
        except HTTPError as err:
            self.postMessage.emit(f"Captioning: send to Microsoft Teams failed with error code {err.code}.")
    def send_zoom(self, cap):
        """Take caption and send post to Zoom session
        """
        req = request.Request(self.endpoint + f"&seq={self.zoom_seq}&lang=en-US", method = "POST")      
        self.zoom_seq += 1
        req.data = cap.encode()  
        req.add_header('Content-Type', 'text/plain')
        req.add_header('Content-Length', len(cap)) 
        try:
            r = request.urlopen(req)
        except HTTPError as err:
            self.postMessage.emit(f"Captioning: send to Zoom failed with error code {err.code}.")  
    def send_obs(self, cap):
        """Take caption and send to OBS using obsws_python
        """
        try:
            res = self.obs.send_stream_caption(cap)
        except OBSSDKRequestError as err:
            self.postMessage.emit(f"Captioning: send to OBS failed with error code {err.code}")
        except Exception as e:
            self.postMessage.emit(f"Captioning: send to OBS failed. Error message is {e}")



import re
import textwrap
from datetime import datetime
from collections import UserList, UserString
from copy import deepcopy
from itertools import accumulate
from bisect import bisect_left, bisect
from plover_cat.helpers import pixel_to_in, write_command
from plover_cat.constants import user_field_dict
from PyQt5.QtCore import QByteArray, QBuffer, QIODevice
from PyQt5.QtGui import QImage, QImageReader
from odf.teletype import addTextToElement
from odf.text import P, UserFieldDecls, UserFieldDecl, UserFieldGet, UserIndexMarkStart, UserIndexMarkEnd
from odf.draw import Frame, TextBox, Image

_whitespace = '\u2029\t\n\x0b\x0c\r '
whitespace = r'[%s]' % re.escape(_whitespace)
wordsep_simple_re = re.compile(r'(%s+)' % whitespace)

class text_element(UserString):
    """The base text element used in editor.

    :param text: string that can be set
    :type text: str
    :param time: time element is created in ISO milliseconds format
    :type time: str
    """
    def __init__(self, text = "", time = None):
        super().__init__(text)
        self.element = "text"
        """type of element, ``text``"""
        self.time = time or datetime.now().isoformat("T", "milliseconds")
    def __len__(self):
        """return length of string"""
        return(len(self.data))
    def split(self):
        """Splits text string on whitespace (re from textwrapper).

        :return: s list of elements containing each text piece separately,
            but same otherwise as original
        :rtype: list
        """
        if self.length() > 1:
            chunks = [c for c in wordsep_simple_re.split(self.data) if c]
            list_chunks = []
            for c in chunks:
                class_dict = deepcopy(self.__dict__)
                class_dict["data"] = c
                new_element = self.__class__()
                new_element.from_dict(class_dict)
                list_chunks.append(new_element)
            return(list_chunks)
        else:
            class_dict = deepcopy(self.__dict__)
            new_element = self.__class__()
            new_element.from_dict(class_dict)
            return([new_element])            
    def __iter__(self):
        """Return element in list length one for iteration. 
        """
        return(iter([self]))
    def __add__(self, other):
        """Adds together text, and updates time from other.
        
        :param other: any type of element
        :type other: ``text_element``
        :return: ``text_element`` object with updated attributes
        :rtype: ``text_element``
        :raise NotImplemented: if type of ``other`` not ``text_element``         
        """
        if type(other) == type(self):
            data = self.data + other.data
            time = other.time
            return(self.__class__(text = data, time = time))
        else:
            return NotImplemented
    def __radd__(self, other):
        """Override __radd__ from ``Userstring`` so inherited classes can override.
        
        :raises NotImplemented: __radd__ should not be needed with purely ``text_element`` objects
        """
        return NotImplemented
    def __getitem__(self, key):
        """Get from class dict based on key.
        
        :param key: key
        :return: returns new instance of class after deepcopy
        """
        class_dict = deepcopy(self.__dict__)
        class_dict["data"] = self.data[key]
        new_element = self.__class__()
        new_element.from_dict(class_dict)
        return(new_element)
    def __repr__(self):
        """Return representation as ``dict``."""
        items = ("%s = %r" % (k, v) for k, v in self.__dict__.items())
        return("{name}({args})".format(name = self.__class__.__name__, args = ", ".join(items)))
    def length(self):
        """Return functional length.
        
        :return: functional length of element, length of text string
        :rtype: int
        """
        return(len(self.data))
    def from_dict(self, dictionary):
        """Populate class using a dict."""
        for k, v in dictionary.items():
            setattr(self, k, v)
    def to_display(self):
        """Formatted string for display in GUI.

        Should be a string for three lines, 1) icon letter, 2) element data, if any, 3) text
        """
        return("\U0001F163\n\n%s" % self.to_text())
    def to_json(self):
        """Return dict of attributes."""
        return(self.__dict__)
    def to_text(self):
        """Return "text" representation as imagined for ``QTextEdit``."""
        return(self.data)
    def to_rtf(self):
        """Return string representation with control groups from RTF/CRE spec as necessary."""
        time_string = datetime.strptime(self.time, "%Y-%m-%dT%H:%M:%S.%f").strftime('%H:%M:%S')      
        string = write_command("cxt", time_string + ":00", visible = False, group = True) + write_command("cxs", "", visible = False, group = True) + self.to_text()
        return(string)
    def to_odt(self, paragraph, document):
        """Populate ODF paragraph element with instance text.
        
        :param paragraph: odfpy paragraph element
        :param document: odfpy document that the paragraph (will) belongs to
        """
        # if there is style highlighting on element level
        # need to general styled text here and add style to document
        addTextToElement(paragraph, self.to_text())
    def replace_initial_tab(self, expand = "    "):
        """Replace first tab in string.
        
        :param expand: string to replace tab with, default four spaces
        :type expand: str
        """
        if "\t" in self.data:
            self.data = self.data.replace("\t", expand, 1)
            return True
        else:
            return None

class dummy_element(text_element):
    """Dummy element used for testing.
    """
    def __init__(self, **kargs):
        super().__init__(**kargs)
        self.element = "dummy"
        """type of element, ```dummy```"""
    def length(self):
        return(1)

class stroke_text(text_element):
    """Stroke element used in editor.

    :param stroke: steno outline, separated by slashes
    :type stroke: str
    :param audiotime: time of media at time of stroke
    :type audiotime: str

    """
    def __init__(self, stroke = "", audiotime = "", **kargs):
        super().__init__(**kargs)
        self.element = "stroke"
        """Type of element, ``stroke``."""
        self.stroke = stroke
        self.audiotime = audiotime
    def __add__(self, other):
        """Adds together stroke elements or stroke and text elements.
        Will only combine elements but not across word boundaries (spaces), 
        can accept `stroke_text` and `text_element` but not others while updating
        necessary attributes.

        :param other: ``stroke_element`` or ``text_element``
        :type other: ``stroke_element`` or ``text_element``
        :return: ``stroke_text`` object with updated attributes
        :rtype: ``stroke_text``
        :raise TypeError: if type of ``other`` not correct 
        """
        if self.data.endswith(" ") or other.data.startswith(" "):
            raise ValueError("Elements should not be combined across word boundaries")
        if type(other) == type(self):
            data = self.data + other.data
            time = other.time
            stroke = self.stroke + "/" + other.stroke
            audiotime = other.audiotime
            return(self.__class__(stroke = stroke, time = time, text = data, audiotime = audiotime))
        elif type(other) == text_element:
            data = self.data + other.data
            time = other.time
            stroke = self.stroke
            audiotime = self.audiotime
            return(self.__class__(stroke = stroke, time = time, text = data, audiotime = audiotime))            
        else:
            raise TypeError("Stroke elements can only combine with other stroke or text elements.")
    def __radd__(self, other):
        """Add together elements.

        :param other: ``text_element`` object
        :return: ``stroke_text`` object with updated attributes
        :rtype: ``stroke_text``
        :raise NotImplemented: if type of ``other`` is not ``text_element``
        :raise ValueError: if trying to combine across word boundaries        
        """
        if self.data.startswith(" ") or other.data.endswith(" "):
            raise ValueError("Elements should not be combined across word boundaries")        
        if isinstance(other, text_element):
            data = other.data + self.data
            time = self.time
            stroke = self.stroke
            audiotime = self.audiotime
            return(self.__class__(stroke = stroke, time = time, text = data, audiotime = audiotime))
        else:
            return NotImplemented
            # raise TypeError("Stroke elements can only combine with other stroke or text elements.")
    def to_rtf(self):
        time_string = datetime.strptime(self.time, "%Y-%m-%dT%H:%M:%S.%f").strftime('%H:%M:%S')      
        string = write_command("cxt", time_string + ":00", visible = False, group = True) + write_command("cxs", self.stroke, visible = False, group = True) + self.data
        return(string)
    def to_display(self):
        return("\U0001F162\n%s\n%s" % (self.stroke, self.data))

class image_text(text_element):
    """Image element used in editor.

    :param path: path to image
    :type path: str or `Path`
    :param width: pixel width of image
    :type width: int
    :param height: pixel height of image
    :type height: int
    """
    def __init__(self, path = None, width = None, height = None, **kargs):
        super().__init__(**kargs)
        self.data = "\ufffc"
        """Representation of image when image cannot be rendered."""
        self.element = "image"
        """Type of element, ``image``."""
        self.path = path
        self.width = width
        self.height = height
    def __add__(self, other):
        """Addition of elements to image is not allowed.
        
        :raise NotImplemented: images cannot add or be added to other elements.
        """
        return NotImplemented
        # raise NotImplementedError("Cannot add on image element.")
    def length(self):
        """Return functional length of ``image_text``, 1.

        :return: 1
        :rtype: int
        """
        return(1)
    def to_display(self):
        return("\U0001F158\n%s\n " % self.path)
    def to_rtf(self):
        image = QImage(QImageReader(self.path).read())
        ba = QByteArray()
        buffer = QBuffer(ba)
        buffer.open(QIODevice.WriteOnly)
        image.save(buffer, "PNG")
        hex_string = ba.toHex()
        string = (write_command("pngblip") + write_command("picw", value = image.width())
                 + write_command("pich", value = image.height()) + write_command("picwgoal", value = image.width() * 15)
                 + write_command("pichgoal", value = image.height() * 15))
        string = write_command("pict", hex_string.data().decode(), value=string, group = True)
        return(string)
    def to_odt(self, paragraph, document):
        width_in = "{0}in".format(pixel_to_in(self.width))
        height_in = "{0}in".format(pixel_to_in(self.height))
        image_frame = Frame(attributes = {"stylename": "Graphics", "anchortype": "as-char", "width": width_in, "height": height_in})
        image_ref = document.addPicture(self.path)
        image_frame.addElement(Image(href = image_ref))
        paragraph.addElement(image_frame)

class text_field(text_element):
    """Field element to use in editor.

    :param name: name of field
    :type name: str
    :param user_dict: transcript field dict, uses default ``user_field_dict`` if not supplied
    :type user_dict: dict
    """
    def __init__(self, name = None, user_dict = user_field_dict, **kargs):
        super().__init__(**kargs)
        self.element = "field"
        """Type of element, ``field``"""
        self.name = name
        self.user_dict = user_dict
        self.update()
    def __add__(self, other):
        """No element can be added. Force use of ``__radd__``.

        :raise NotImplemented: addition not allowed 
        """
        return NotImplemented
        # raise NotImplementedError("Cannot add on text field element.")
    def length(self):
        """Return functional length, 1

        :return: 1
        :rtype: int
        """
        return(1)
    def to_json(self):
        new = {k: v for k, v in self.__dict__.items() if k != "user_dict"}
        return(new)        
    def to_display(self):
        self.update()
        return("\U0001F155\n \n%s" % self.data)
    def to_text(self):
        self.update()
        return(self.data)
    def update(self):
        """Update data for display. Check the ``user_dict`` and assigns value to ``data``, 
        "updating" if the value in the dict for the ``name`` has been changed.
        The ``user_field_dict`` by default uses the global `user_field_dict` object.
        Need to specify if a different dict is to be used.
        """
        if self.name in self.user_dict:
            self.data = self.user_dict[self.name]
        else:
            self.data = "{%s}" % str(self.name)
    def to_rtf(self):
        self.update()
        rtf_string = write_command("fldinst", "DOCVARIABLE %s" % self.name, visible=False, group=True)
        rtf_string = rtf_string + write_command("fldrslt", self.data)
        rtf_string = write_command("fldlock", rtf_string, group = True)
        return(write_command("field", rtf_string))
    def to_odt(self, paragraph, document):
        self.update()
        # move this out to actual function, since this is setting document var
        user_declares = document.text.getElementsByType(UserFieldDecls)
        if user_declares:
            field_names = [i.getAttribute("name") for i in document.text.getElementsByType(UserFieldDecl)]
            if not self.name in field_names:
                user_declares = document.text.getElementsByType(UserFieldDecls)[0]
                user_dec = UserFieldDecl(name = self.name, stringvalue = self.data, valuetype = "string")
                user_declares.addElement(user_dec)                
        else:
            user_declares = UserFieldDecls()
            user_dec = UserFieldDecl(name = self.name, stringvalue = self.data, valuetype = "string")
            user_declares.addElement(user_dec)
            document.text.addElement(user_declares)
        user_field = UserFieldGet(name = self.name)
        user_field.addText(self.data)
        paragraph.addElement(user_field)

class automatic_text(stroke_text):
    """Automatic text element for editor, text added by editor, not user.
    
    Use for text such as Q\t, ie set directly for question style.
    
    :param prefix: text to appear before string
    :type prefix: str
    :param suffix: text to appear after string
    :type suffix: str
    """
    def __init__(self, prefix = "", suffix = "", **kargs):
        super().__init__(**kargs)
        self.element = "automatic"
        """Type of element, ``automatic``"""
        self.prefix = prefix
        self.suffix = suffix
    def __add__(self, other):
        """No element can be added. Force use of ``__radd__``.

        :raise NotImplemented: addition not allowed.
        """
        return NotImplemented
    # raise NotImplementedError("Cannot add on automatic text element.")
    def __radd__(self, other):
        """No element can be added.

        :raise NotImplemented: ``__radd__`` not allowed.
        """
        return NotImplemented
    def __len__(self):
        """Return length of prefix + text + suffix."""
        return(len(self.to_text()))
    def to_text(self):
        return(self.prefix + self.data + self.suffix)
    def length(self):
        """Return length of data, compare to ``__len__()``."""
        return(len(self.data))
    def to_rtf(self):
        string = ""
        if self.prefix:
            string = string + write_command("cxa",  self.prefix, visible = False, group = True)
        string = string + write_command("cxt", self.time + ":00", visible = False, group = True) + write_command("cxs", self.stroke, visible = False, group = True) + self.data
        if self.suffix:
            string = string + write_command("cxa",  self.suffix, visible = False, group = True)
        return(string)
    def to_display(self):
        return(f"\U0001F162 \U0001F150\n{self.stroke}\n{self.to_text()}")

class conflict_text(stroke_text):
    """Not yet implemented"""
    # need for resolving with imports from rtf
    def __init__(self, choices = None, **kargs):
        super().__init__(**kargs)
        self.choices = choices

class index_text(text_element):
    """Index entry element for editor.

    :param prefix: prefix to place before "number"
    :type prefix: str
    :param indexname: index that index entry belongs to
    :type indexname: int
    :param description: index entry description
    :type description: str
    :param hidden: whether description should be shown in editor or hidden
    :type hidden: bool
    """
    def __init__(self, prefix = "Exhibit", indexname = 0, description = "", hidden = True, **kargs):
        super().__init__(**kargs)
        self.element = "index"
        """Type of element, ``index``"""
        # indexname and data ("text") are identifiers, not allowed to change
        self.indexname = indexname
        self.prefix = prefix
        self.description = description
        self.hidden = hidden
    def __add__(self, other):
        """No element can be added. Force use of ``__radd__``.

        :raise NotImplemented: addition not allowed.
        """        
        return NotImplemented
        # raise NotImplementedError("Cannot add on index text element.")
    def __len__(self):
        """Return length of index entry text, may or may not include description."""
        return(len(self.to_text()))
    def length(self):
        """Return functional length, 1

        :return: 1
        :rtype: int
        """
        return(1)
    def to_text(self):
        if self.hidden:
            return(f"{self.prefix}\u0020{self.data}")
        else:
            return(f"{self.prefix}\u0020{self.data}{self.description}")
    def to_display(self):
        return(f"\U0001F154\n\n{self.to_text()}")
    def to_odt(self, paragraph, document):
        # while first instinct is to use self.time, problem with copied steno if index
        id_time = str(datetime.now().timestamp())
        index_start = UserIndexMarkStart(id = id_time, indexname = self.indexname)
        index_end = UserIndexMarkEnd(id = id_time)
        paragraph.addElement(index_start)
        addTextToElement(paragraph, self.to_text())
        paragraph.addElement(index_end)
    def to_rtf(self):
        # Exhibit 1{\xe\cxinum1\v Exhibit 1:  A knife}
        # {\xe{\*\cxexnum 1}Exhibit 1:  A knife}
        string = write_command("cxinum", value = self.indexname)
        if self.hidden:
            string = string + write_command("v")
        string = write_command("cxexnum", self.data, visible = False, group = True) + string
        string = write_command("xe", text = f"{self.prefix}\u0020{self.data}{self.description}", value = string, group = True)
        if self.hidden:
            string = self.to_text() + string
        return(string)

class redact_text(text_element):
    """Not yet implemented"""
    def __init__(self, **kargs):
        super().__init__(**kargs)
        self.element = "redacted"
    def to_display(self):
        return("\U0001F161\n\n%s" % self.data)
    def to_text(self):
        return("\u2588" * len(self.data))

def translate_coords(len1, len2, pos):
    """Translate position from one sequence of cumulative lengths to another.
    
    :param len1: cumulative text lengths of elements (ie in display of editor)
    :type len1: list[int]
    :param len2: cumulative functional lengths of elements, can be 1 or other int
    :type len2: list[int]
    :param pos: position from first sequence to translate
    :type pos: int
    
    :return: position in second sequence
    :rtype: int
    """
    pos_index = bisect_left(len1, pos)
    if pos_index == 0:
        remainder = pos     
    elif pos in len1:
        remainder = 0
    else:
        remainder = pos - len1[pos_index - 1]
    if remainder == 0:
        steno_pos = len2[pos_index]
    else:
        if remainder > (len2[pos_index] - len2[pos_index - 1]):
            steno_pos = len2[pos_index - 1] + (len2[pos_index] - len2[pos_index - 1])
        else:
            steno_pos = len2[pos_index - 1] + remainder  
    return(steno_pos) 

def backtrack_coord(pos, backspace, text_len, func_len):
    """Return position after backspace, accounting for functional length of elements.

    :param pos: initial position in text
    :type pos: int
    :param backspace: hypothetical backspaces to mock
    :type backspace: int
    :param text_len: text lengths of elements
    :type text_len: list[int]
    :param func_len: functional lengths of elements
    :type func_len: list[int]

    :return: position after hypothetical backspaces, may be negative
        if more ``backspaces`` than length
    :rtype: int
    """
    # guard against that edgecase
    if backspace == 0:
        return(pos)
    cum_text_len = list(accumulate(text_len))
    cum_text_len.insert(0, 0)
    cum_func_len = list(accumulate(func_len))
    cum_func_len.insert(0, 0)
    func_pos = translate_coords(cum_text_len, cum_func_len, pos)
    ending_func_pos = func_pos - backspace
    if ending_func_pos < 0:
        # if more backspaces than exists in collection
        return(ending_func_pos)
    cum_func_index = bisect_left(cum_func_len, ending_func_pos)
    if cum_func_index == 0:
        return(0)
    elif ending_func_pos in cum_func_len:
        return(cum_text_len[cum_func_index])
    else:
        return(cum_text_len[cum_func_index - 1] +  ending_func_pos - cum_func_len[cum_func_index - 1])

# text_len = [4, 6, 4, 6, 5, 5]
# func_len = [4, 1, 4, 1, 5, 5]
# pos = 25
# backspace = 5
# backtrack_coord(pos, backspace, text_len, func_len)
# backtrack_coord(24, backspace, text_len, func_len)
# text_len = [4, 3, 3, 5]
# func_len = [4, 3, 1, 5]
# pos = 10
# backspace = 1
# backtrack_coord(pos, backspace, text_len, func_len)

class element_factory:
    """Factory for creating elements from data dict"""
    def gen_element(self, element_dict, user_field_dict = user_field_dict):
        """Return element based on type.
        
        :param element_dict: dict of element, likely from dict representation
        :type element_dict: dict
        :param user_field_dict: user field data
        :type user_field_dict: dict
        :return: element
        :rtype: `text_element` or subclass
        """
        # default is always a text element
        element = text_element()
        if element_dict["element"] == "stroke":
            element = stroke_text()
        # elif element_dict["element"] == "dummy":
        #     element = dummy_element()
        elif element_dict["element"] == "image":
            element = image_text()
        elif element_dict["element"] == "field":
            element = text_field(user_dict = user_field_dict)
        elif element_dict["element"] == "automatic":
            element = automatic_text()
        elif element_dict["element"] == "index":
            element = index_text()
        element.from_dict(element_dict)
        return(element)

class element_collection(UserList):
    """Container for holding elements in list."""
    def __init__(self, data = None):
        # force element into list if not list
        if isinstance(data, list):
            super().__init__(data)
        elif not data:
            super().__init__([])
        else:
            super().__init__([data])
    def __iter__(self):
        return(iter(self.data))
    def __str__(self):
        """Return string representation of all elements in container."""
        string = [i.to_text() for i in self.data]
        return("".join(string))
    def lengths(self):
        """Return list of functional lengths for each element."""
        lengths = [i.length() for i in self.data]
        return(lengths)
    def lens(self):
        """Return list of "text" lengths for each element."""
        lens = [len(i) for i in self.data]
        return(lens)
    def __len__(self):
        """Returns sum of `len` for each element."""
        return(sum(self.lens()))
    def __getitem__(self, key):
        """Return `element_collection` instance with copy of element(s) based on key."""
        if isinstance(key, slice):
            el_lengths = list(accumulate(self.lengths()))
            start = key.start
            if not start:
                start = 0
            end = key.stop
            el_part = self.__class__()
            if end == 0:
                return(el_part)
            if not self.data:
                return(el_part)
            if not end:
                end = el_lengths[-1]
            if start < 0 or end < 0:
                raise ValueError("negative slices not supported")            
            if end > max(el_lengths):
                raise IndexError('list index out of range')
            if start == el_lengths[-1]:
                return(el_part)
            first_whole = bisect_left(el_lengths, start)
            if first_whole == 0:
                first_remain = start
            else:
                first_remain = start - el_lengths[(first_whole - 1)]
            last_whole = bisect_left(el_lengths, end)
            if last_whole == 0:
                last_remain = end
            else:
                last_remain = end - el_lengths[(last_whole - 1)]
            # special case where first and last are within same element
            data = deepcopy(self.data)
            if first_whole == last_whole:
                el_part.append(data[last_whole][first_remain:last_remain])
                return(el_part)
            if not start in el_lengths:
                el_part.append(data[first_whole][first_remain:])
            if (first_whole + 1) != last_whole:
                for i in data[(first_whole + 1): last_whole]:
                    el_part.append(i)
            if not end in el_lengths:
                el_part.append(data[last_whole][:last_remain])
            else:
                el_part.append(data[last_whole])
            return(el_part)
        else:
            return(self.__class__(deepcopy(self.data[key])))
        # always return element collection, even when not a slice
    def element_count(self):
        """Return number of elements in collection, ``len`` of a list."""
        return(len(self.data))
    def to_json(self):
        """Return list of serialized ``dict`` objects."""
        return([i.to_json() for i in self.data])
    def to_text(self):
        """Return text string combining all elements."""
        text = [i.to_text() for i in self.data]
        return("".join(text))
    def to_rtf(self):
        """Return string containing RTF representations of elements."""
        col_string = "".join([i.to_rtf() for i in self.data])
        return(col_string)
    def to_odt(self, paragraph, document):
        """Add each element to paragraph in ODF document"""
        for i in self.data:
            i.to_odt(paragraph, document)
    def to_display(self):
        """Return list of display strings for elements"""
        return([el.to_display() for el in self.data])
    def to_strokes(self):
        """Return strin with all strokes"""
        el_strokes = [el.stroke for el in self.data if el.element == "stroke"]
        return("/".join(el_strokes))
    def remove(self, start, end):
        """Remove elements based on specified functional position start/stop.

        :param int start: start position for remove
        :param int end: end position for remove
        :return: removed elements
        :rtype: ``element_collection``
        """
        new_data = []
        first = self.__getitem__(slice(0, start)).data
        new_data.extend(first)
        second = self.__getitem__(slice(end, None)).data
        new_data.extend(second)
        del_data = self.__getitem__(slice(start, end))
        self.data = new_data
        return(del_data)
    def insert(self, i, item):
        """Insert based on functional position.
        :param int i: position
        :param item: data to be inserted
        :return: item
        """
        new_data = []
        first = self.__getitem__(slice(0, i))
        second = self.__getitem__(slice(i, None))
        new_data.extend(first.data)
        if isinstance(item, UserList):
            new_data.extend(item.data)
        else:
            new_data.append(item)
        new_data.extend(second.data)
        self.data = new_data
        return(item)
    def stroke_pos_at_pos(self, pos):
        """Returns tuple of text start, stop for element at text ``pos``."""
        lengths = self.lens()
        cum_len = list(accumulate(lengths))
        pos_index = bisect(cum_len, pos)
        # if first, need manual adjustment
        if pos_index == 0:
            start_pos = 0
        else:
            start_pos = cum_len[pos_index - 1]
        # if last, pos_index will cause out of range error, subtract back 
        if pos >= cum_len[-1]:
            pos_index = pos_index - 1
            start_pos = cum_len[pos_index - 1]
        return((start_pos, cum_len[pos_index]))
    def element_pos(self, index):
        """Returns tuple of text start, stop for element at ``index`` in collection."""
        lengths = self.lens()
        cum_len = list(accumulate(lengths))
        if index == 0:
            start_pos = 0
        else:
            start_pos = cum_len[index - 1]
        if index >= len(cum_len):
            index = index - 1
            start_pos = cum_len[index - 1]            
        return((start_pos, cum_len[index]))            
    def remove_steno(self, start, end):
        """Removes elements from `start` to `end` (text) position.
        
        :return: element(s) removed
        :rtype: ``element_collection``
        """
        lens = self.lens()
        cum_len = list(accumulate(lens))
        cum_len.insert(0, 0)
        # print(cum_len)
        lengths = self.lengths()
        cum_lengths = list(accumulate(lengths))
        cum_lengths.insert(0, 0)
        start_pos = translate_coords(cum_len, cum_lengths, start)
        end_pos = translate_coords(cum_len, cum_lengths, end)
        res = self.remove(start_pos, end_pos)
        return(res)
    def extract_steno(self, start, end):
        """Retrieve elements from ``start`` to ``end`` (text) position.

        This does not modify original collection.

        :param int start: starting text position
        :param int end: ending text position
        :return: element(s) between coordinates
        :rtype: ``element_collection``
        """
        lens = self.lens()
        cum_len = list(accumulate(lens))
        cum_len.insert(0, 0)
        # print(cum_len)
        lengths = self.lengths()
        cum_lengths = list(accumulate(lengths))
        cum_lengths.insert(0, 0)
        start_pos = translate_coords(cum_len, cum_lengths, start)
        end_pos = translate_coords(cum_len, cum_lengths, end)
        res = self[start_pos:end_pos]
        return(res)        
    def insert_steno(self, pos, item):
        """Insert at text position.

        This will split an element in collection if needed.

        :param int pos: text position
        :param item: ``element_collection`` or single element
        :return: item
        """
        lens = self.lens()
        cum_len = list(accumulate(lens))
        cum_len.insert(0, 0)
        # print(cum_len)
        lengths = self.lengths()
        cum_lengths = list(accumulate(lengths))
        cum_lengths.insert(0, 0)
        steno_pos = translate_coords(cum_len, cum_lengths, pos)
        res = self.insert(steno_pos, item)
        return(res)
    def starts_with(self, char):
        """Check if text starts with ``char``.

        :param str char: text to check
        :return: ``True`` if collection text does start with ``char``, 
            ``False`` otherwise
        :rtype: bool
        """
        return(self.data[0].data.startswith(char))
    def ends_with(self, char):
        """Check if text ends with ``char``

        :param str char: text to check
        :return: ``True if collection text ends with ``char``,
            ``False`` otherwise
        :rtype: bool
        """
        return(self.data[-1].data.endswith(char))
    def starts_with_element(self, element_type):
        """Check if first element is of type.
        
        :param str element_type: any of the types for ``text_element`` or subclasses
        :return: ``True`` if first element type matches ``element_type``, ``False`` otherwise
        :rtype: bool
        """
        if not self.data:
            return False
        if self.data[0].element == element_type:
            return True
        else:
            return False
    def ends_with_element(self, element_type):
        """Check if last element is of type.
        
        :param str element_type: any of the types for ``text_element`` or subclasses
        :return: ``True`` if last element type matches ``element_type``, ``False`` otherwise
        :rtype: bool
        """
        if self.data[-1].element == element_type:
            return True
        else:
            return False
    def remove_end(self, char = "\n"):
        """Remove ``char`` end of text string for collection.

        This will remove entire element if ``char`` is only text in element.

        :param str char: string to remove, default ``\\n``
        """
        if self.data[-1].data == char:
            del self.data[-1]
        elif self.data[-1].data.endswith(char):
            self.data[-1].data = self.data[-1].data.rstrip(char) 
    def remove_begin(self, char):
        """Remove ``char`` from first element if text starts with ``char``.

        :param str char: string to remove
        """
        if self.data[0].data == char:
            del self.data[0]
        elif self.data[0].data.startswith(char):
            self.data[0].data = self.data[0].data.lstrip(char)
    def add_begin(self, char = " "):
        """Add ``char`` to beginning of first element.
        
        :param str char: string to add, default one space character
        """
        self.data[0].data = char + self.data[0].data
    def add_end(self, char = " "):
        """Add ``char`` to end of last element.
        
        :param str char: string to add, default one space character
        """
        self.data[-1].data = self.data[-1].data + char
    def stroke_count(self):
        """Counts the number of strokes in collection."""
        # for RTF, maybe has uses elsewhere
        return(sum([el.stroke.count("/") + 1 for el in self.data if el.element == "stroke"]))
    def search_strokes(self, query):
        """Return text positions for matches to underlying strokes.

        :param str query: steno outline
        :return: tuple of start and end positions, ``None`` if no match
        """
        stroke_list = [el.stroke if el.element == "stroke" else " " for el in self.data]
        query = query.split("/")
        # must match across strokes, match whole element of stroke
        # cannot match across stroke auto_text stroke
        match = False
        for i, subsets in enumerate(zip(*(stroke_list[i:] for i in range(len(query))))):
            if tuple(query) == subsets:
                match = True
                break
        if not match:
            return None
        start_pos = self.element_pos(i)[0]
        end_pos = self.element_pos(i + len(query))[0]    
        return((start_pos, end_pos))
    def search_text(self, query):
        """Return text positions for matches to text.

        :param str query: search text
        :return: results of ``re.finditer`` search
        """        
        text = "".join([el.to_text() for el in self.data])
        res = re.finditer(re.escape(query), text)
        return(res)
    def collection_time(self, reverse = False):
        """Return earliest/latest timestamp in collection.

        :param bool reverse: ``True`` by default for earliest, 
            ``False`` for latest
        :return: formatted timestamp string
        """
        times = [el.time for el in self.data]
        return(sorted(times, reverse = reverse)[0])
    def audio_time(self, reverse = False):
        """Return earliest/latest audio timestamp in collection.

        :param bool reverse: ``True`` by default for earliest, 
            ``False`` for latest
        :return: formatted timestamp string
        """
        times = [el.audiotime for el in self.data if el.element == "stroke" and el.audiotime != ""]
        if times:
            return(sorted(times, reverse = reverse)[0])
        else:
            return None
    def replace_initial_tab(self, tab_replace = "    "):
        """Replace initial tab in place within collection.

        :param str tab_replace: string to replace tab character with, default four spaces
        """
        track_len = 3
        for el in self.data:
            if "\t" in el.data[0:track_len]:
                res = el.replace_initial_tab(tab_replace)
                if res:
                    break
            track_len -= len(el)
            if track_len < 0:
                break
    def merge_elements(self):
        """Collapse collection elements using ``__add__`` method.
        """
        new_ec = []
        last_el_type = ""
        for ind, el in enumerate(self.data):
            if ind == 0:
                new_ec.append(el)
                continue
            try:
                sum_el = new_ec[-1] + el
                new_ec[-1] = sum_el
            except (TypeError, ValueError, NotImplementedError):
                new_ec.append(el)
        return(self.__class__(new_ec))

# stroke_data = [text_element(text = "ABC"), stroke_text(stroke = "T-", text = "it "), text_element(text = "2 ", time = "2023-08-09T23:02:26.526"), text_element(text = "3 "), stroke_text(stroke = "EUFS ", text = "I was "), stroke_text(stroke = "TAO", text = "too ")]
# ex_text = index_text(description = "index descript", text = "index name")
# stroke_data.append(ex_text)

# stroke_collection = element_collection(stroke_data)
# stroke_collection.merge_elements()
# stroke_collection.replace_initial_tab("  ")
# element_list = stroke_collection.to_json() 
# el_factory = element_factory()
# [el_factory.gen_element(element_dict = i) for i in element_list]
# stroke_collection.search_strokes("T-")
# stroke_collection
# ex_text.data = "oops"
# str(stroke_collection)
# stroke_collection.stroke_pos_at_pos(2)
# stroke_collection.stroke_pos_at_pos(7)
# stroke_collection.stroke_pos_at_pos(15)
# stroke_collection[0]
# stroke_collection[2]
# stroke_collection[6]
# stroke_collection[2:7]
# # stroke_collection.remove(2,7)
# new_stroke_data = [text_element(text = "ABCD"), text_element(text = "123"), text_element(text = "456")]

# stroke_collection.insert(2, new_stroke_data)
# new_stroke_collection = element_collection(new_stroke_data)

# # stroke_collection.insert(5, new_stroke_collection)
# stroke_collection.insert_steno(7, new_stroke_collection)
# str(stroke_collection)

class steno_wrapper(textwrap.TextWrapper):
    """Wrap text, but adapted for elements in ``element_collection``.
    
    :return: a list of lists of elements, not ``element_collection``.
    """
    def __init__(self, **kargs):
        super().__init__(**kargs)

    def _split(self, text):
        # override
        text.remove_end()
        merged = text.merge_elements()
        chunks = []
        for el in merged:
            chunks.extend(el.split())
        return chunks

    def _wrap_chunks(self, chunks):
        lines = []
        if self.width <= 0:
            raise ValueError("invalid width %r (must be > 0)" % self.width)
        if self.max_lines is not None:
            if self.max_lines > 1:
                indent = self.subsequent_indent
            else:
                indent = self.initial_indent
            if len(indent) + len(self.placeholder.lstrip()) > self.width:
                raise ValueError("placeholder too large for max width")

        chunks.reverse()

        while chunks:
            cur_line = []
            cur_len = 0
            if lines:
                indent = self.subsequent_indent
            else:
                indent = self.initial_indent
            width = self.width - len(indent)
            # if self.drop_whitespace and chunks[-1].strip() == '' and lines:
            #     del chunks[-1]
            while chunks:
                l = len(chunks[-1])
                if cur_len + l <= width:
                    cur_line.append(chunks.pop())
                    cur_len += l
                else:
                    break
            if chunks and len(chunks[-1]) > width:
                self._handle_long_word(chunks, cur_line, cur_len, width)
                cur_len = sum(map(len, cur_line))

        #     # If the last chunk on this line is all whitespace, drop it.
        #     if self.drop_whitespace and cur_line and cur_line[-1].strip() == '':
        #         cur_len -= len(cur_line[-1])
        #         del cur_line[-1]

            if cur_line:
                lines.append(cur_line)
        return lines

    def _split_chunks(self, text):
        return self._split(text)

    def wrap(self, text):
        chunks = self._split_chunks(text)
        return self._wrap_chunks(chunks)

# wrap_text = steno_wrapper(width = 10)
# wrap_text.wrap(stroke_collection)


from PyQt5.QtWidgets import QDialog, QTableWidgetItem, QMessageBox
from PyQt5.QtCore import Qt, pyqtSignal
from collections import Counter
from plover import log
from plover.steno import normalize_steno
from plover_cat.helpers import extract_ngram
from plover_cat.constants import stopwords
from plover_cat.suggest_dialog_ui import Ui_suggestDialog

class suggestDialogWindow(QDialog, Ui_suggestDialog):
    """Analyze existing transcript for common phrases and words.

    The suggest dialog presently works by analyzing the open transcript,
    through other sources are possible in the future. It can do a word frequency
    search, an n-gram search, or both, filtering by minimum occurrence. The
    word frequency search can use the SCOWL size filter, 
    while very common words are filtered out first.
    
    :param text: string of text
    :type text: str
    :param engine: existing Plover engine
    :param scowl_dict: SCOWL word list packaged with Plover
    :type scowl_dict: dict
    """
    # insert_autocomplete = pyqtSignal(tuple)
    def __init__(self, text, engine, scowl_dict):
        """Sets up connections for analysis with dialog UI"""
        super().__init__()
        self.setupUi(self)  
        self.document = text
        self.engine = engine
        self.detect.clicked.connect(self.analyze)
        self.scowl_dict = scowl_dict
        # self.toAutocomplete.clicked.connect(self.send_autocomplete)
        self.toDictionary.clicked.connect(self.send_dictionary)
        self.displaySuggest.setColumnCount(3)
        self.displaySuggest.setHorizontalHeaderLabels(["Candidate", "Outline", "Alternative outlines"])
    def analyze(self):
        """Starts analyzing text based on selected filters and parameters."""
        search_type = self.searchType.currentText()
        if search_type == "Words only":
            result_list = self.analyze_words(self.scowlSize.currentText(), self.minOccur.value())
        elif search_type == "N-grams only":
            result_list = self.analyze_ngrams(self.minNgram.value(), self.maxNgram.value(), self.minOccur.value())
        else:
            n_list = self.analyze_ngrams(self.minNgram.value(), self.maxNgram.value(), self.minOccur.value())
            word_list = self.analyze_words(self.scowlSize.currentText(), self.minOccur.value())
            result_list = n_list + word_list
        self.displaySuggest.clear()
        self.displaySuggest.setRowCount(0)
        self.displaySuggest.setColumnCount(3)
        self.displaySuggest.setHorizontalHeaderLabels(["Candidate", "Outline", "Alternative outlines"])
        self.displaySuggest.setRowCount(len(result_list))
        for row, word in enumerate(result_list):
            self.displaySuggest.setItem(row, 0, QTableWidgetItem(word))
            outlines = self.get_outline(word)
            if outlines:
                self.displaySuggest.setItem(row, 1, QTableWidgetItem(outlines[0][0]))
                if len(outlines) > 1:
                    alternatives = ", ".join([out[0] for out in outlines])
                    self.displaySuggest.setItem(row, 2, QTableWidgetItem(alternatives))
            else:
                self.displaySuggest.setItem(row, 1, QTableWidgetItem(""))
                self.displaySuggest.setItem(row, 2, QTableWidgetItem(""))
    def analyze_ngrams(self, min_ngram = 2, max_ngram = 3, min_occurrence = 3):
        """Run n-gram search and filter for display."""
        log.debug("Running ngram search.")
        if max_ngram < min_ngram:
            max_ngram = min_ngram + 1
        word_counter = Counter()
        for n in range(min_ngram, max_ngram + 1):
            for par in self.document.splitlines():
                # print([" ".join(i) for i in list(extract_ngram(par, n))])
                word_counter.update(list(extract_ngram(par, n)))
        result_list = [" ".join(list(k)) for k, v in word_counter.items() if v >= min_occurrence]
        return(result_list)
    def analyze_words(self, scowl_size = 35,  min_occurrence = 1):
        """Run word search and filter based on occurrence and SCOWL size."""
        log.debug("Running word frequency search.")
        word_counter = Counter(self.document.split())
        common = result_list = [k for k, v in word_counter.items() if v >= min_occurrence]
        result_list = []
        if scowl_size == "":
            for ind, word in enumerate(common):
                if word.lower() not in stopwords:
                    result_list.append(word)                    
        else:
            scowl_size = int(scowl_size)
            for ind, word in enumerate(common):
                if word in self.scowl_dict and self.scowl_dict[word] > scowl_size:
                    result_list.append(word)
                elif word not in self.scowl_dict:
                    result_list.append(word)
        return(result_list)
    def get_outline(self, translation):
        """Get list of outlines for translation from engine."""
        # one day, lookup strokes from rtf or other transcripts
        res = self.engine.reverse_lookup(translation)
        return(list(res))
    def update_text(self, text):
        """Update text for analysis"""
        self.document = text
    # def send_autocomplete(self):
    #     selected_row = self.displaySuggest.currentRow()
    #     if selected_row == -1:
    #         QMessageBox.warning(self, "Add to autocomplete", "No row selected.")
    #         return
    #     word = self.displaySuggest.item(selected_row, 0).text()
    #     stroke = self.displaySuggest.item(selected_row, 1).text()
    #     self.insert_autocomplete.emit((word, stroke))
    def send_dictionary(self):
        """Add selected outline to dictionary."""
        selected_row = self.displaySuggest.currentRow()
        if selected_row == -1:
            QMessageBox.warning(self, "Plover2CAT", "No row selected for adding to dictionary.")
            return
        word = self.displaySuggest.item(selected_row, 0).text()
        stroke = self.displaySuggest.item(selected_row, 1).text()
        if len(stroke) == 0:
            QMessageBox.warning(self, "Plover2CAT", "Missing translation outline to add to dictionary.")    
            return
        self.engine.add_translation(normalize_steno(stroke, strict = True), word.strip())


import unittest
import pathlib
import os
from tempfile import mkdtemp, mkstemp
from shutil import rmtree
from io import StringIO
from unittest import TextTestRunner

from PyQt5.QtWidgets import QDialog, QListWidgetItem
from PyQt5.QtGui import QTextCursor
from PyQt5.QtCore import Qt
from plover.oslayer.config import CONFIG_DIR
from plover.steno import Stroke, normalize_stroke, normalize_steno
from plover_cat.helpers import save_json
from plover_cat.steno_objects import *
from plover_cat.TextEditor import PloverCATEditor
from plover_cat.test_dialog_ui import Ui_testDialog

class TestStenoData(unittest.TestCase):
    def test_text_element_creation(self):
        el = text_element(text = "abc")
        self.assertEqual(el.element, "text")
        self.assertEqual(el.to_text(), "abc")
        self.assertEqual(len(el), 3)
        self.assertEqual(el.length(), 3)
    def test_steno_element_creation(self):
        el = stroke_text(stroke = "T-", text = "it")
        self.assertEqual(el.element, "stroke")
        self.assertEqual(el.to_text(), "it")
        self.assertEqual(len(el), 2)
        self.assertEqual(el.length(), 2)
    def test_field_element_creation(self):
        el = text_field(name = "SPEAKER_STPHAO")
        val = "Mr. Stphao"
        self.assertEqual(el.element, "field")
        self.assertEqual(el.to_text(), val)
        self.assertEqual(len(el), len(val))
        self.assertEqual(el.length(), 1)
        self.assertNotEqual(el.length, len(val))
    def test_image_element(self):
        el = image_text()
        self.assertEqual(el.element, "image")
        self.assertEqual(len(el), 1)        
        self.assertEqual(el.length(), 1)                
    def test_automatic_element(self):
        txt = "ABC"
        affix_txt = "Q.\tABC?"
        el = automatic_text(prefix = "Q.\t", suffix = "?", stroke = "AEU", text = txt)
        self.assertEqual(el.element, "automatic")
        self.assertEqual(el.to_text(), affix_txt)
        self.assertNotEqual(len(el), len(txt))
        self.assertEqual(len(el), len(affix_txt))
        self.assertEqual(el.length(), len(txt))
        self.assertNotEqual(el.length(), len(affix_txt))
    def test_index_element(self):
        indices = {"0": {"prefix": "Exhibit", "hidden": True, "entries": {"A": " A glass.", "B": " Two glasses."}},
                     "1": {"prefix": "Index", "hidden": False, "entries": {"1": " The first.", "2": " The second."}}}
        index_name = "0"
        index_dict = indices[index_name]
        entry_text = "A"
        correct_txt = "Exhibit A"
        el = index_text(prefix = index_dict["prefix"], indexname = index_name, description = index_dict["entries"][entry_text], hidden = index_dict["hidden"], text = entry_text)
        self.assertEqual(el.element, "index")
        self.assertEqual(el.to_text(), correct_txt)
        self.assertEqual(len(el), len(correct_txt))
        self.assertNotEqual(len(el), 1)
        self.assertEqual(el.length(), 1)
        self.assertNotEqual(el.length(), len(correct_txt))
        index_name = "1"
        index_dict = indices[index_name]
        entry_text = "1"
        correct_txt = "Index 1 The first."
        el = index_text(prefix = index_dict["prefix"], indexname = index_name, description = index_dict["entries"][entry_text], hidden = index_dict["hidden"], text = entry_text)
        self.assertEqual(el.to_text(), correct_txt)
        self.assertEqual(len(el), len(correct_txt))
        self.assertNotEqual(len(el), 1)
        self.assertEqual(el.length(), 1)
        self.assertNotEqual(el.length(), len(correct_txt))
    def element_addition(self):
        el1 = text_element("ABC")
        el2 = text_element("DEF")
        res = el1 + el2 
        self.assertEqual(res.to_text(), "ABCDEF")
        el_stroke = stroke_text(stroke = "T-", text = " it")
        res = el1 + el_stroke
        self.assertEqual(res.to_text(), "ABC it")
        res = el_stroke + el1 
        self.assertEqual(res.to_text(), " itABC")
        self.assertEqual(res.stroke, "T-")
        el_stroke2 = stroke_text(stroke = "-B", text = "be")
        res = el_stroke + el_stroke2
        self.assertEqual(res.to_text(), " itbe")
        self.assertEqual(res.stroke, "T-/-B")
        with self.assertRaises(ValueError):
            el_stroke2 + el_stroke
        el_index = index_text(description = "index descript", text = " index name")
        with self.assertRaises(NotImplemented):
            el_stroke + el_index
            el_index + el_stroke
    def test_element_collection(self):
        stroke_data = [text_element(text = "ABC"),  
                        text_element(text = "2 ", time = "2000-01-23T00:00:00.111"), 
                        stroke_text(stroke = "EUFS", text = "I was ", time = "2100-01-24T00:00:00.111"), 
                        stroke_text(stroke = "TAO", text = "too "), 
                        index_text(description = "index descript", text = " index name")]
        sc = element_collection(stroke_data)
        self.assertEqual(sc.element_count(), 5)
        self.assertEqual(sc.stroke_count(), 2)
        self.assertEqual(sc.collection_time(), "2000-01-23T00:00:00.111")
        self.assertEqual(sc.collection_time(reverse = True), "2100-01-24T00:00:00.111")
        merged = sc.merge_elements()
        self.assertEqual(merged.element_count(), 4)
        sc.remove_steno(15, 16)
        self.assertEqual(sc.to_text(), "ABC2 I was too ")
        sc.remove_steno(3, 6)
        self.assertEqual(sc.to_text(), "ABC was too ")
        new_data = [text_element(text = "ABCD"), text_element(text = "123")]
        sc.insert_steno(4, element_collection(new_data))
        self.assertEqual(sc.to_text(), "ABC ABCD123was too ")

class TestTextEdit(unittest.TestCase):
    def __init__(self, testname, editor, selection):
        super(TestTextEdit, self).__init__(testname)
        self.editor = editor
        self.temp_dir = None
        self.output_setting = False
        self.space_setting = "Before Output"
        self.selection = selection
    def setUp(self):
        self.temp_dir = mkdtemp()
        self.output_setting = self.editor.engine.output
        self.editor.engine.set_output(True)
        self.space_setting = self.editor.engine.config["space_placement"]
        self.editor.engine.config["space_placement"] = "Before Output"
        self.step_Open()
    def tearDown(self):
        self.editor.engine.set_output(self.output_setting)
        self.editor.engine.config["space_placement"] = self.space_setting
        self.editor.textEdit.undo_stack.setClean()
        self.editor.close_file()
        rmtree(self.temp_dir)
    def testMonolithic(self):
        for _s in self.selection:
            try:
                getattr(self, _s)()
            except Exception as e:
                self.fail('{} failed({})'.format(_s, e))   
    def step_Open(self):
        file_path = pathlib.Path(self.temp_dir) / "test"
        self.editor.open_file(file_path)
    def step_ConfirmFiles(self):
        engine_dicts = self.editor.engine.config["dictionaries"]
        default_dict_dir = self.editor.textEdit.file_name / "dict"
        self.assertTrue(default_dict_dir.exists())
        default_dict = default_dict_dir / "default.json"
        self.assertTrue(default_dict.exists())
        style_file = self.editor.textEdit.file_name / "styles" / "default.json"
        self.assertTrue(style_file.exists())
    def step_Stroke(self):
        self.editor.engine.clear_translator_state()
        # test append
        self.editor.on_send_string("THE")
        self.assertEqual(self.editor.textEdit.last_string_sent, "THE")
        self.editor.textEdit.on_stroke(Stroke("-T"))
        self.assertEqual(self.editor.textEdit.toPlainText().strip(), "THE")
        stroke_data = self.editor.textEdit.textCursor().block().userData()["strokes"]
        self.assertEqual(stroke_data.element_count(), 1)
        # test delete
        self.editor.count_backspaces(3)
        self.editor.textEdit.on_stroke(Stroke("*"))
        self.assertEqual(self.editor.textEdit.toPlainText(), "")
        self.assertEqual(stroke_data.element_count(), 0)
        # test log
        self.editor.textEdit.log_to_tape(Stroke("*"))
        self.assertNotEqual(len(self.editor.strokeList.toPlainText()), 0)
        self.editor.textEdit.clear_transcript()
    def step_WriteTwoLine(self):
        # todo: create mock tape to read instead
        self.editor.on_send_string("ABC\\nDEF")
        self.editor.textEdit.on_stroke(Stroke("S"))
        self.assertEqual(self.editor.textEdit.toPlainText(), "ABC\\nDEF")
        self.editor.textEdit.clear_transcript()        
    def step_AddRemoveDict(self):
        dict_dir = self.editor.textEdit.file_name / "dict"
        handle, new_dict = mkstemp(suffix = ".json")
        new_dict_contents = {"S": "Success"}
        save_json(new_dict_contents, new_dict)
        self.editor.add_dict(new_dict)
        self.assertEqual(len([p for p in dict_dir.iterdir() if p.suffix == ".json"]), 2)
        self.assertEqual(self.editor.engine.lookup(tuple("S")), "Success")
        new_dict_path = dict_dir / pathlib.Path(new_dict).name
        self.editor.remove_dict(new_dict_path)
        self.assertEqual(len([p for p in dict_dir.iterdir() if p.suffix == ".json"]), 2)
        self.assertNotEqual(self.editor.engine.lookup(tuple("S")), "Success")
        os.close(handle)
        pathlib.Path(new_dict).unlink()
    def step_SplitPar(self):
        one_text = {0: {"style": "Normal", "strokes": [{"data": "ABCDEF", "element": "stroke", "stroke": "S-", "time": "2000-01-01T00:00:00.001"}]}}
        self.editor.textEdit.undo_stack.setClean()
        save_json(one_text, self.editor.textEdit.file_name.joinpath(self.editor.textEdit.file_name.stem).with_suffix(".transcript"))
        self.editor.close_file()
        self.editor.open_file(pathlib.Path(self.temp_dir) / "test")
        cursor = self.editor.textEdit.textCursor()
        cursor.setPosition(3)
        self.editor.textEdit.setTextCursor(cursor)
        self.editor.textEdit.split_paragraph()
        self.assertEqual(self.editor.textEdit.toPlainText(), "ABC\nDEF")
        self.editor.textEdit.undo_stack.undo()
        self.assertEqual(self.editor.textEdit.toPlainText(), "ABCDEF")
        stroke_data = self.editor.textEdit.textCursor().block().userData()["strokes"]
        self.assertEqual(len(stroke_data.data), 1)
    def step_SplitParSpace(self):
        one_text = {0: {"style": "Normal", "strokes": [{"data": "ABC DEF", "element": "stroke", "stroke": "S-", "time": "2000-01-01T00:00:00.001"}]}}
        self.editor.textEdit.undo_stack.setClean()
        save_json(one_text, self.editor.textEdit.file_name.joinpath(self.editor.textEdit.file_name.stem).with_suffix(".transcript"))
        self.editor.close_file()
        self.editor.open_file(pathlib.Path(self.temp_dir) / "test")
        cursor = self.editor.textEdit.textCursor()
        cursor.setPosition(3)
        self.editor.textEdit.setTextCursor(cursor)
        self.editor.textEdit.split_paragraph()
        self.assertEqual(self.editor.textEdit.toPlainText(), "ABC\nDEF")
        self.editor.textEdit.undo_stack.undo()
        cursor.setPosition(3)
        self.editor.textEdit.setTextCursor(cursor)
        self.editor.textEdit.split_paragraph(remove_space = False)
        self.assertEqual(self.editor.textEdit.toPlainText(), "ABC\n DEF")
    def step_MergePar(self):
        one_text = {0: {"style": "Normal", "strokes": [{"data": "ABC", "element": "stroke", "stroke": "S-", "time": "2000-01-01T00:00:00.001"},
                                                         {"data": "\n", "element": "stroke", "stroke": "R-R", "time": "2000-01-01T00:00:00.002"}]},
                    1: {"style": "Normal", "strokes": [{"data": "DEF", "element": "stroke", "stroke": "-T", "time": "2000-01-01T00:00:00.002"}]}}
        self.editor.textEdit.undo_stack.setClean()
        save_json(one_text, self.editor.textEdit.file_name.joinpath(self.editor.textEdit.file_name.stem).with_suffix(".transcript"))
        self.editor.close_file()
        self.editor.open_file(pathlib.Path(self.temp_dir) / "test")  
        cursor = self.editor.textEdit.textCursor()
        cursor.setPosition(3)
        self.editor.textEdit.setTextCursor(cursor)
        self.editor.textEdit.merge_paragraphs()   
        self.assertEqual(self.editor.textEdit.toPlainText(), "ABC\nDEF")
        cursor.setPosition(5) 
        self.editor.textEdit.setTextCursor(cursor)
        self.editor.textEdit.merge_paragraphs()
        self.assertEqual(self.editor.textEdit.toPlainText(), "ABC DEF")
        self.editor.textEdit.undo_stack.undo()
        self.assertEqual(self.editor.textEdit.toPlainText(), "ABC\nDEF")
    def step_CheckStyleAttr(self):
        one_text = {0: {"style": "Normal", "strokes": [{"data": "ABC", "element": "stroke", "stroke": "S-", "time": "2000-01-01T00:00:00.001"},
                                                         {"data": "\n", "element": "stroke", "stroke": "R-R", "time": "2000-01-01T00:00:00.002"}]},
                    1: {"style": "Question", "strokes": [{"data": "DEF", "element": "stroke", "stroke": "-T", "time": "2000-01-01T00:00:00.002"}]}}
        self.editor.textEdit.undo_stack.setClean()
        save_json(one_text, self.editor.textEdit.file_name.joinpath(self.editor.textEdit.file_name.stem).with_suffix(".transcript"))
        self.editor.close_file()
        self.editor.open_file(pathlib.Path(self.temp_dir) / "test")  
        cursor = self.editor.textEdit.textCursor()
        cursor.setPosition(2)
        self.assertEqual(cursor.block().userData()["style"], "Normal")
        cursor.setPosition(5)
        self.assertEqual(cursor.block().userData()["style"], "Question")
    def step_ChangeStyle(self):
        one_text = {0: {"style": "Normal", "strokes": [{"data": "ABC", "element": "stroke", "stroke": "S-", "time": "2000-01-01T00:00:00.001"},
                                                         {"data": "\n", "element": "stroke", "stroke": "R-R", "time": "2000-01-01T00:00:00.002"}]},
                    1: {"style": "Question", "strokes": [{"data": "DEF", "element": "stroke", "stroke": "-T", "time": "2000-01-01T00:00:00.002"}]}}
        self.editor.textEdit.undo_stack.setClean()
        save_json(one_text, self.editor.textEdit.file_name.joinpath(self.editor.textEdit.file_name.stem).with_suffix(".transcript"))
        self.editor.close_file()
        self.editor.open_file(pathlib.Path(self.temp_dir) / "test")  
        cursor = self.editor.textEdit.textCursor()
        cursor.movePosition(5)
        self.editor.textEdit.setTextCursor(cursor)
        self.editor.textEdit.set_paragraph_style("Normal")
        self.assertEqual(cursor.block().blockFormat().textIndent(), 0)
        self.editor.textEdit.set_paragraph_style("Question")
        self.assertEqual(cursor.block().blockFormat().textIndent(), 0.5 * 96)
        tabs = [tab.position for tab in cursor.block().blockFormat().tabPositions()]
        self.assertEqual(96 in tabs, True)
    def step_loadNewStyle(self):
        pass
    def step_ColorHighlight(self):
        # set color, close transcript, re-open, check color
        # then insert text directly in transcript
        self.editor.textEdit.clear_transcript()         

class testDialogWindow(QDialog, Ui_testDialog):
    """Dialog to run selected tests for editor and transcript.
    """
    def __init__(self, editor):
        super().__init__()
        self.setupUi(self)
        self.editor = editor
        self.runTest.clicked.connect(self.run_tests)
        self.selectAll.clicked.connect(self.select_all)
        self.unselectAll.clicked.connect(self.unselect_all)
        self.selection = {"step_ConfirmFiles": "Setup files are present", 
                    "step_Stroke": "Strokes create output and are logged", 
                    "step_AddRemoveDict": "Add and remove new dict",
                    "step_WriteTwoLine": "Write correctly with new line in translation",
                    "step_SplitPar": "Split paragraph and undo",
                    "step_SplitParSpace": "Split paragraph with space involved and undo",
                    "step_MergePar": "Merge paragraph, space involved, and undo",
                    "step_CheckStyleAttr": "Text properly styled when loaded",
                    "step_ChangeStyle": "Text properly styled when style changed manually"}
        last = len(self.selection)
        counter = 0
        for i, des in self.selection.items():
            item = QListWidgetItem()
            item.setText(des)
            item.setData(Qt.UserRole, i)
            item.setCheckState(Qt.Unchecked)
            counter += 1
            if counter == last:
                item.setCheckState(Qt.Checked)
            self.listSelection.addItem(item)
    def select_all(self):
        for index in range(0, self.listSelection.count()):
            item = self.listSelection.item(index)
            item.setCheckState(Qt.Checked)
    def unselect_all(self):
        for index in range(0, self.listSelection.count()):
            item = self.listSelection.item(index)
            item.setCheckState(Qt.Unchecked)        
    def display_log(self):
        log_path = pathlib.Path(CONFIG_DIR) / "plover.log"
        log_contents = log_path.read_text()
        self.output.clear()
        self.output.setPlainText(log_contents)
        self.output.moveCursor(QTextCursor.End)
    def run_tests(self):
        res_output = StringIO()
        suite_steno = unittest.TestLoader().loadTestsFromTestCase(TestStenoData)
        suite = unittest.TestSuite([suite_steno])
        selection = []
        for index in range(0, self.listSelection.count()):
            item = self.listSelection.item(index)
            if item.checkState() == Qt.Checked:
                selection.append(item.data(Qt.UserRole))
        suite.addTest(TestTextEdit("testMonolithic", self.editor, selection))
        # suite.addTest(TestTextEdit("testConfirmFiles", self.editor))
        res = TextTestRunner(stream = res_output, verbosity=1).run(suite)
        self.output.clear()
        self.output.setPlainText(res_output.getvalue())



from copy import deepcopy
from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtWidgets import QTableWidgetItem, QDialog
from plover_cat.index_dialog_ui import Ui_indexDialog
from plover_cat.steno_objects import index_text

class indexDialogWindow(QDialog, Ui_indexDialog):
    """Create, save and edit index entries for editor.

    The index editor dialog is a non-modal dialog opened through the menu.
    Index entries are numbered, and have a description that can be hidden.
    Index entries have to belong to an index with a prefix string, and there 
    can be multiple indices.

    :param index_dict: dict with {index_name: {prefix: str, hidden: bool, entries: {name: description, ...}}, ..}
    :type index_dict: dict    
    """
    index_insert = pyqtSignal(object)
    """Signal emitted when user has selected an entry to insert into text"""
    updated_dict = pyqtSignal()
    """Signal emitted when the index dictionary has been updated"""
    def __init__(self, index_dict):
        super().__init__()
        self.setupUi(self)
        self.index_dict = index_dict
        self.indexChoice.addItems([*self.index_dict])
        self.displayEntries.setColumnCount(2)
        self.displayEntries.setHorizontalHeaderLabels(["Name", "Description"])
        self.displayEntries.itemSelectionChanged.connect(self.enable_insert)
        self.indexChoice.activated.connect(self.update_display)
        self.indexChoice.currentIndexChanged.connect(self.enable_add)
        self.saveIndex.clicked.connect(self.save_index)
        self.saveAndInsert.clicked.connect(self.insert_entry)
        self.addNewIndex.clicked.connect(self.new_index)
        self.entryAdd.clicked.connect(self.add_new_entry)
    def update_display(self, index):
        """Display existing entries for selected index."""
        self.indexChoice.setCurrentIndex(index)
        self.indexPrefix.clear()
        if not self.hideDescript.isChecked():
            self.hideDescript.setChecked(True)
        self.displayEntries.clear()
        self.displayEntries.setRowCount(0)
        self.entryText.clear()
        index_name = self.indexChoice.currentText()
        if index_name not in self.index_dict:
            return
        self.indexPrefix.setText(self.index_dict[index_name]["prefix"])
        if self.index_dict[index_name]["hidden"] != self.hideDescript.isChecked():
            self.hideDescript.setChecked(self.index_dict[index_name]["hidden"])
        self.displayEntries.setRowCount(len(self.index_dict[index_name]["entries"]))
        self.displayEntries.setHorizontalHeaderLabels(["Name", "Description"])
        for row, (k, v) in enumerate(self.index_dict[index_name]["entries"].items()):
            key = QTableWidgetItem(k)
            key.setFlags(key.flags() & ~Qt.ItemIsEditable & ~ Qt.ItemIsUserCheckable)
            self.displayEntries.setItem(row, 0, key)
            self.displayEntries.setItem(row, 1, QTableWidgetItem(v))
    def save_index(self):
        """Save selected index."""
        index_name = self.indexChoice.currentText()
        prefix = self.indexPrefix.text()
        hidden = self.hideDescript.isChecked()
        entry_dict = {}
        for row in range(self.displayEntries.rowCount()):
            key = self.displayEntries.item(row, 0).text()
            description = self.displayEntries.item(row, 1).text()
            entry_dict[key] = description
        self.index_dict[index_name] = {"prefix": prefix, "hidden": hidden, "entries": entry_dict}
        self.updated_dict.emit()
    def new_index(self):
        """Create new index."""
        count = self.indexChoice.count()
        self.indexChoice.addItem(str(count))
        self.update_display(count)
    def enable_add(self, index):
        """Allow new entry to be added if index is selected."""
        if index > -1:
            self.entryAdd.setEnabled(True)
    def add_new_entry(self):
        """Add new entry to currently selected index."""
        index_name = self.indexChoice.currentText()
        new_entry = self.entryText.text()
        prefix = self.indexPrefix.text()
        if index_name not in self.index_dict:
            self.index_dict[index_name] = {"prefix": prefix, "hidden": True, "entries": {}}
        self.index_dict[index_name]["entries"][new_entry] = ""
        self.update_display(self.indexChoice.currentIndex())
    def enable_insert(self):
        """Allow insert into text if an entry is selected."""
        if self.displayEntries.currentRow() > -1:
            self.saveAndInsert.setEnabled(True)
    def insert_entry(self):
        """Insert selected entry into text"""
        index_name = self.indexChoice.currentText()
        selected_row = self.displayEntries.currentRow()
        if selected_row == -1:
            return
        entry_text = self.displayEntries.item(selected_row, 0).text()
        index_dict = self.index_dict[index_name]
        el = index_text(prefix = index_dict["prefix"], indexname = index_name, description = index_dict["entries"][entry_text], hidden = index_dict["hidden"], text = entry_text)
        self.index_insert.emit(el)
        self.save_index()
    def update_dict(self, index_dict):
        """Update existing index dict, used when indices are gathered from transcript """
        self.index_dict = index_dict
        self.indexChoice.clear()
        self.indexChoice.addItems([*self.index_dict])
        if len(self.index_dict) > 0:
            self.update_display(0)
        else:
            self.update_display(-1)



import bisect
import re 
from PyQt5.QtGui import QTextBlockFormat, QFont, QTextCharFormat, QTextOption
from PyQt5.QtCore import Qt
from plover_cat.steno_objects import *
from plover_cat.helpers import *
from copy import deepcopy
from odf.style import (Style, TextProperties, ParagraphProperties, FontFace, PageLayout, 
PageLayoutProperties, MasterPage, TabStops, TabStop, GraphicProperties, Header, Footer)

def format_odf_text(block_data, style, chars_in_inch, page_width, line_num = 0):
    """Format a string into wrapped lines for ODF.

    This uses ``steno_wrap_ODF`` and does additional formatting to wrapped
    lines by converting necessary styling parameters.

    :param block_data: an ``element_collection``
    :param dict style: styling parameters in dict
    :param int chars_in_inch: approximate number of characters in inch for selected style font
    :param page_width: page width of ODF document, in inches
    :param int line_num: line number for starting line
    :return: dict of dicts ``{line_number: {line_text, line_timestamp}}``
    """
    # block is QTextBlock
    # max_char fed into function is converted from page width or user setting (line num digits and timestamp digits are "outside of page")
    text = block_data.to_text()
    l_marg = 0
    r_marg = 0
    first_indent = 0
    # block_data = deepcopy(block.userData()["strokes"])
    spaces_to_insert = 0
    if "paragraphproperties" in style:
        if "marginleft" in style["paragraphproperties"]:
            l_marg = inch_to_spaces(style["paragraphproperties"]["marginleft"], chars_in_inch)
        if "marginright" in style["paragraphproperties"]:
            r_marg = inch_to_spaces(style["paragraphproperties"]["marginright"], chars_in_inch)
        if "textindent" in style["paragraphproperties"]:
            first_indent = inch_to_spaces(style["paragraphproperties"]["textindent"], chars_in_inch)
            spaces_to_insert = first_indent
    max_char = inch_to_spaces(page_width, chars_in_inch) - l_marg - r_marg
    # print(max_char)
    # this only deals with a tab at the start of the paragraph
    if "\t" in text[0:3]:
        mat = re.search("\t", text)
        begin_pos = first_indent + l_marg + mat.start()
        # print(begin_pos)
        if "tabstop" in style["paragraphproperties"]:
        # example with 0.5 par indent, 0.5 text indent
        # original string: Q.\tABC
        # formatted:
        # ----------Q.\tABC
        # tabstop at 1.5in
        # ----------Q.---ABC
        # ---------------{start}
            tabs = style["paragraphproperties"]["tabstop"]
            if isinstance(tabs, list):
                tabs = [inch_to_spaces(i, chars_in_inch) for i in tabs]
                tab_index = bisect.bisect_left(tabs, begin_pos)
                tabs = tabs[tab_index]
            else:
                tabs = inch_to_spaces(tabs, chars_in_inch)
        else:
            # just use the python default tab expansion value
            tabs = 8
        if first_indent > 0:
            spaces_to_insert = (tabs - begin_pos) + first_indent
        else:
            spaces_to_insert = (tabs - begin_pos)
    par_text = steno_wrap_odf(block_data = block_data, max_char = max_char,
                                tab_space=4, first_line_indent=" " * spaces_to_insert, par_indent = "", starting_line_num=line_num)
    # print(par_text)
    for k, v in par_text.items():
        # erase whitespace at front since odf formatting is separate
        par_text[k]["text"][0].data = par_text[k]["text"][0].data.lstrip(" ")
        # each item line should end with \n except last
        if not par_text[k]["text"][-1].endswith("\n"):
            par_text[k]["text"][-1].data = par_text[k]["text"][-1].data + ("\n")
    if list(par_text.keys()):
        # last line of paragraph should not have a "\n" char, new paragraph is done through odf elements
        par_text[list(par_text.keys())[-1]]["text"][-1].data = par_text[list(par_text.keys())[-1]]["text"][-1].data.rstrip("\n")
    return(par_text)

def format_text(block_data, style, max_char = 80, line_num = 0):
    """Format a string into wrapped lines for text.

    This uses ``steno_wrap_plain`` and does additional formatting to wrapped
    lines by converting styling parameters to padding space characters to make
    formatted text lines.

    :param block_data: an ``element_collection``
    :param dict style: styling parameters in dict
    :param int max_char: maximum characters in a line
    :param int line_num: line number for starting line
    :return: dict of dicts ``{line_number: {line_text, line_timestamp}}``
    """
    # block is QTextBlock
    # max_char fed into function is converted from page width or user setting (line num digits and timestamp digits are "outside of page")
    text = block_data.to_text()
    l_marg = 0
    r_marg = 0
    t_marg = 0
    b_marg = 0
    align = "left"
    linespacing = "100%"
    first_indent = 0
    # print(block_data)
    # all measurements converted to spaces, for horizontal, 10 chars per inch, for vertical, 6 chars per inch
    if "paragraphproperties" in style:
        if "marginleft" in style["paragraphproperties"]:
            l_marg = inch_to_spaces(style["paragraphproperties"]["marginleft"])
        if "marginright" in style["paragraphproperties"]:
            r_marg = inch_to_spaces(style["paragraphproperties"]["marginright"])
        if "margintop" in style["paragraphproperties"]:
            t_marg = inch_to_spaces(style["paragraphproperties"]["margintop"], 6)
        if "marginbottom" in style["paragraphproperties"]:
            b_margin = inch_to_spaces(style["paragraphproperties"]["marginbottom"], 6)
        if "linespacing" in style["paragraphproperties"]:
            linespacing = style["paragraphproperties"]["linespacing"]
        if "textalign" in style["paragraphproperties"]:
            align = style["paragraphproperties"]["textalign"]
            # do not support justified text for now
            if align == "justify":
                align = "left"
        if "textindent" in style["paragraphproperties"]:
            first_indent = inch_to_spaces(style["paragraphproperties"]["textindent"])
    # since plaintext, none of the textproperties apply
    max_char = max_char - l_marg - r_marg
    # this only deals with a tab at the start of the paragraph
    if "\t" in text[0:3] and "tabstop" in style["paragraphproperties"]:
        # example with 0.5 par indent, 0.5 text indent
        # original string: Q.\tABC
        # formatted:
        # ----------Q.\tABC
        # tabstop at 1.5in
        # ----------Q.---ABC
        # ---------------{start}
        tabs = style["paragraphproperties"]["tabstop"]
        mat = re.search("\t", text)
        begin_pos = first_indent + l_marg + mat.start()
        if isinstance(tabs, list):
            tabs = [inch_to_spaces(i) for i in tabs]
            tab_index = bisect.bisect_left(tabs, begin_pos)
            tabs = tabs[tab_index]
        else:
            tabs = inch_to_spaces(tabs)
        # tabs_to_space_pos = inch_to_spaces(tabs)
        spaces_to_insert = (tabs - begin_pos) * " "
        text = re.sub("\t", spaces_to_insert, text, count = 1)
        block_data.replace_initial_tab(spaces_to_insert)
    par_text = steno_wrap_plain(text = text, block_data = block_data, max_char = max_char,
                                tab_space=4, first_line_indent=" " * (first_indent + l_marg), par_indent = " " * l_marg, starting_line_num=line_num)
    for k, v in par_text.items():
        if align == "center":
            par_text[k]["text"] = par_text[k]["text"].center(max_char)
        elif align == "right":
            par_text[k]["text"] = par_text[k]["text"].rjust(max_char)
        else:
            par_text[k]["text"] = par_text[k]["text"].ljust(max_char)
    if t_marg > 0:
        par_text[line_num+1]["text"] = "\n" * t_marg + par_text[line_num+1]["text"]
    if b_marg > 0:
        par_text[-1]["text"] = par_text[-1]["text"] + "\n" * b_marg
    # decide whether to add extra line(s) due to linespacing, will round to two extra empty lines if over 149%
    line_spaces = int(int(linespacing.replace("%", "")) / 100) - 1
    for k, v in par_text.items():
        par_text[k]["text"] = par_text[k]["text"] + "\n" * line_spaces
    return(par_text)

def format_srt_text(block_data, line_num = 0, audiostarttime = "", audioendtime = ""):
    """Wrap steno data for SRT.

    :param block_data: an ``element_collection`` for a paragraph
    :param int line_num: starting line number
    :param audiostarttime: paragraph's ``audiostarttime``
    :param audioendtime: paragraph's ``audioendtime``
    :return: dict of dicts ``{line_number: {line_text, line_timestamp}}``
    """
    par_text = steno_wrap_srt(block_data, max_char = 47, starting_line_num = line_num)
    # line_times = []
    # if audiostarttime and audioendtime:
    #     for line, data in par_txt.items():
    #         line_times.append(data["starttime"])
    #         line_times.append(data["endtime"])
    #     if all([t >= audiostarttime and t <= audioendtime for t in line_times]):
    #         return(par_text)
    #     else:
    #         audiostarttime = hours_to_ms(audiostarttime)
    #         audioendtime = hours_to_ms(audioendtime)
    #         # not complete
    return(par_text)

def steno_wrap_plain(text, block_data, max_char = 80, tab_space = 4, first_line_indent = "", 
                        par_indent = "", starting_line_num = 0):
    """Wrap steno data properly for text.
    
    See code to compare for differences with ``steno_wrap_odf`` and ``steno_wrap_srt``.

    :param str text: text to wrap
    :param block_data: an ``element_collection`` for a paragraph
    :param int max_char: maximum number of characters per line
    :param int tab_space: how many spaces should a tab character be converted into
    :param str first_line_indent: string to place at beginning of first line
    :param str par_indent: string to place at beginning of every line except first
    :param int starting_line_num: line number to add to beginning of each line
    :return: dict of dicts ``{line_number: {line_text, line_timestamp}}``
    
    """
    # the -1 in max char is because the rounding is not perfect, might have some lines that just tip over
    # uses text string instead of block_data because text has pre-expanded tabs 
    wrapped = textwrap.wrap(text, width = max_char - 1, initial_indent= first_line_indent,
                subsequent_indent= par_indent, expand_tabs = False, tabsize = tab_space, replace_whitespace=False)
    begin_pos = 0
    par_dict = {}
    for ind, i in enumerate(wrapped):
        matches = block_data.extract_steno(begin_pos, len(block_data)).search_text(i.strip())
        # matches = re.finditer(re.escape(i.strip()), text[begin_pos:])
        for i in matches:
            if i.start() >= 0:
                match = i
                break
        end_pos = match.end()
        ## try very desperately to recover
        if end_pos == begin_pos:
            end_pos = begin_pos + len(wrapped[ind]) - 1
        line_time = block_data.extract_steno(begin_pos, end_pos).collection_time()
        begin_pos = match.end()
        par_dict[starting_line_num + ind + 1] = {"text": wrapped[ind], "time": line_time}
    return(par_dict)

def steno_wrap_odf(block_data, max_char = 80, tab_space = 4, first_line_indent = "", 
                        par_indent = "", starting_line_num = 0):
    """Wrap steno data properly for ODF.

    Compare with ``steno_wrap_plain``.
    
    :param block_data: an ``element_collection`` for a paragraph
    :param int max_char: maximum number of characters per line
    :param int tab_space: how many spaces should a tab character be converted into
    :param str first_line_indent: string to place at beginning of first line
    :param str par_indent: string to place at beginning of every line except first
    :param int starting_line_num: line number to add to beginning of each line
    :return: dict of dicts ``{line_number: {line_text, line_timestamp}}``
    
    """
    # the -1 in max char is because the rounding is not perfect, might have some lines that just tip over
    wrapper = steno_wrapper(width = max_char - 1, initial_indent= first_line_indent,
                subsequent_indent= par_indent, expand_tabs = False, tabsize = tab_space, replace_whitespace=False)
    wrapped = wrapper.wrap(text = block_data)
    begin_pos = 0
    par_dict = {}
    for ind, i in enumerate(wrapped):
        line_time = element_collection(i).collection_time()
        par_dict[starting_line_num + ind + 1] = {"text": wrapped[ind], "time": line_time}
    return(par_dict)

def steno_wrap_srt(block_data, max_char = 47, tab_space = 0, first_line_indent = "", 
                        par_indent = "", starting_line_num = 0):
    """Wrap steno data properly, but with audio timestamps.

    Compare to ``steno_wrap_odf``
    
    :param block_data: an ``element_collection`` for a paragraph
    :param int max_char: maximum number of characters per line
    :param int tab_space: how many spaces should a tab character be converted into
    :param str first_line_indent: string to place at beginning of first line
    :param str par_indent: string to place at beginning of every line except first
    :param int starting_line_num: line number to add to beginning of each line
    :return: dict of dicts ``{line_number: {line_text, starttime, endtime}}``
    
    """                        
    # change from other wrapper here, tabs are expanded into 0 spaces
    wrapper = steno_wrapper(width = max_char - 1, initial_indent= first_line_indent,
                subsequent_indent= par_indent, expand_tabs = True, tabsize = tab_space, replace_whitespace=False)
    wrapped = wrapper.wrap(text = block_data)
    begin_pos = 0
    par_dict = {}
    for ind, i in enumerate(wrapped):
        ec = element_collection(i)
        start_time = ec.audio_time()
        end_time = ec.audio_time(reverse = True)
        par_dict[starting_line_num + ind + 1] = {"text": wrapped[ind], "starttime": start_time, "endtime": end_time}
    return(par_dict)

def load_odf_styles(path):
    """Extract styles from ODT file and convert supported parameters to par and text style dicts."""
    log.debug(f"Loading ODF style file from {str(path)}")
    style_text = load(path)
    json_styles = {}
    for style in style_text.getElementsByType(Styles)[0].getElementsByType(Style):
        if style.getAttribute("family") != "paragraph":
             continue
        json_styles[style.getAttribute("name")] = {"family": style.getAttribute("family"), "nextstylename": style.getAttribute("nextstylename"), "defaultoutlinelevel": style.getAttribute("defaultoutlinelevel"), "parentstylename": style.getAttribute("parentstylename")}
        if style.getElementsByType(ParagraphProperties):
            par_prop = style.getElementsByType(ParagraphProperties)[0]
            par_dict = {"textalign": par_prop.getAttribute("textalign"), "textindent": par_prop.getAttribute("textindent"), 
                        "marginleft": par_prop.getAttribute("marginleft"), "marginright": par_prop.getAttribute("marginright"), 
                        "margintop": par_prop.getAttribute("margintop"), "marginbottom": par_prop.getAttribute("marginbottom"), "linespacing": par_prop.getAttribute("linespacing")}
            if par_prop.getElementsByType(TabStops):
                tabstop = []
                for i in par_prop.getElementsByType(TabStop):
                    tabstop.append(i.getAttribute("position"))
                par_dict["tabstop"] = tabstop
            json_styles[style.getAttribute("name")]["paragraphproperties"] = par_dict
        if style.getElementsByType(TextProperties):
            txt_prop = style.getElementsByType(TextProperties)[0]
            txt_dict = {"fontname": txt_prop.getAttribute("fontname"), "fontfamily": txt_prop.getAttribute("fontfamily"), 
                        "fontsize": txt_prop.getAttribute("fontsize"), "fontweight": txt_prop.getAttribute("fontweight"), 
                        "fontstyle": txt_prop.getAttribute("fontstyle"), "textunderlinetype": txt_prop.getAttribute("textunderlinetype"), 
                        "textunderlinestyle": txt_prop.getAttribute("textunderlinestyle")}
            json_styles[style.getAttribute("name")]["textproperties"] = txt_dict
    json_styles = remove_empty_from_dict(json_styles)
    return(json_styles)

def recursive_style_format(style_dict, style, prop = "paragraphproperties"):
    """Get full style par/text format dict if style inherits from another."""
    if "parentstylename" in style_dict[style]:
        parentstyle = recursive_style_format(style_dict, style_dict[style]["parentstylename"], prop = prop)
        if prop in style_dict[style]:
            parentstyle.update(style_dict[style][prop])
        return(deepcopy(parentstyle))
    else:
        if prop in style_dict[style]:
            return(deepcopy(style_dict[style][prop]))
        else:
            return({})

def parprop_to_blockformat(par_dict):
    """Take dict of paragraph attributes, return ``QTextBlockFormat``."""
    par_format = QTextBlockFormat()
    if "textalign" in par_dict:
        if par_dict["textalign"] == "justify":
            par_format.setAlignment(Qt.AlignJustify)
        elif par_dict["textalign"] == "right":
            par_format.setAlignment(Qt.AlignRight)
        elif par_dict["textalign"] == "center":
            par_format.setAlignment(Qt.AlignHCenter)
        else:
            # set default to left
            par_format.setAlignment(Qt.AlignLeft)
    if "textindent" in par_dict:
        par_format.setTextIndent(in_to_pixel(par_dict["textindent"].replace("in","")))
    if "marginleft" in par_dict:
        par_format.setLeftMargin(in_to_pixel(par_dict["marginleft"].replace("in", "")))
    if "marginright" in par_dict:
        par_format.setRightMargin(in_to_pixel(par_dict["marginright"].replace("in", "")))
    if "marginbottom" in par_dict:
        par_format.setBottomMargin(in_to_pixel(par_dict["marginbottom"].replace("in", "")))
    if "linespacing" in par_dict:
        par_format.setLineHeight(
            float(par_dict["linespacing"].replace("%", "")),
            QTextBlockFormat.ProportionalHeight
        )
    if "tabstop" in par_dict:
        tab_list = par_dict["tabstop"]
        blocktabs = []
        if isinstance(tab_list, str):
            tab_list = [tab_list]
        for i in tab_list:
            tab_pos = in_to_pixel(i.replace("in", ""))
            blocktabs.append(QTextOption.Tab(tab_pos, QTextOption.LeftTab))
        par_format.setTabPositions(blocktabs)
    return(par_format)

def txtprop_to_textformat(txt_dict):
    """Take dict of text attributes, return ``QTextCharFormat``."""
    txt_format = QTextCharFormat()
    if "fontfamily" in txt_dict:
        potential_font = QFont(txt_dict["fontfamily"])
    else:
        # if no font, then set a default
        potential_font = QFont("Courier New")
    if "fontsize" in txt_dict:
        potential_font.setPointSize(int(float(txt_dict["fontsize"].replace("pt", ""))))
    else:
        # must have a font size
        potential_font.setPointSize(12)
    if "fontweight" in txt_dict and txt_dict["fontweight"] == "bold":
        # boolean for now, odf has weights and so does qt
        potential_font.setBold(True)
    if "fontstyle" in txt_dict and txt_dict["fontstyle"] == "italic":
        potential_font.setItalic(True)
    if "textunderlinetype" in txt_dict and txt_dict["textunderlinestyle"] == "solid":
        potential_font.setUnderline(True)
    txt_format.setFont(potential_font, QTextCharFormat.FontPropertiesAll)
    return(txt_format)




import collections
import time
import pathlib
from shutil import copyfile
from plover import log
from PyQt5.QtGui import QTextBlockUserData, QTextDocument, QTextCursor, QTextBlock, QImage, QImageReader, QTextImageFormat
from PyQt5.QtCore import QUrl, QVariant
from PyQt5.QtWidgets import QUndoCommand
from datetime import datetime, timezone
from plover_cat.steno_objects import *

class BlockUserData(QTextBlockUserData):
    """Representation of the data for a block.
    
    This was adapted from ninja-ide by using a default dict as ``attrs``
    in the class. An empty ``element_collection`` is set in ``self.attr["strokes"]``
    so every block will have an ``element_collection`` set.
    """
    def __init__(self):
        QTextBlockUserData.__init__(self)
        self.attrs = collections.defaultdict(str)
        self.attrs["strokes"] = element_collection()
    def get(self, name, default=None):
        return self.attrs.get(name, default)
    def __getitem__(self, name):
        return self.attrs[name]
    def __setitem__(self, name, value):
        self.attrs[name] = value
    def return_all(self):
        return self.attrs
    def __len__(self):
        return len(self.attrs)

def update_user_data(block_dict, key, value = None):
    """Update BlockUserData key with default value (time) if not provided in a new instance.

    :param block_dict: a ``BlockUserData`` from a ``QTextBlock``
    :param str key: attribute to be updated
    :value str: value of attribute, automatically insert time if ``None``
    :return: updated ``BlockUserData``

    """
    old_dict = block_dict.return_all()
    new_dict = BlockUserData()
    for k, v in old_dict.items():
        new_dict[k] = v
    if value is None:
        value = datetime.now().isoformat("T", "milliseconds")
    new_dict[key] = value
    return new_dict 

class element_actions:
    """QCommand factory for mass insertions.
    
    This is necessary because images are inserted differently than other "text" elements.

    :param document: reference to ``QTextEdit`` instance
    :param int block: block number
    :param int position_in_block: position within block 
    :param element: element to be inserted
    :return: ``QUndoCommand`` that should be pushed into a ``QUndoStack``
    """
    def make_action(self, document, block, position_in_block, element):
        current_cursor = document.textCursor()
        if element.element == "image":
            cmd = image_insert(current_cursor, document, block, position_in_block, element)
        else:
            # treat all "text" elements the same
            cmd = steno_insert(current_cursor, document, block, position_in_block, element)
        return(cmd)

class steno_insert(QUndoCommand):
    """Inserts text and steno data into editor.

    :param cursor: a ``QTextCursor`` instance
    :param document: a ``QTextDocument`` to act upon
    :param int block: ``blockNumber`` of the ``QTextDocument`` to act upon
    :param int position_in_block: position in the block to act upon
    :param steno: elements or ``element_collection`` to insert

    """
    def __init__(self, cursor, document, block, position_in_block, steno):
        super().__init__()
        self.document = document        
        self.block = block
        self.position_in_block = position_in_block
        self.steno = steno
        self.block_state = 1
        self.cursor = cursor
    def redo(self):
        current_cursor = self.cursor
        if current_cursor.blockNumber() == self.block:
            current_block = current_cursor.block()
        else:
            current_block = self.document.document().findBlockByNumber(self.block)
        current_cursor.setPosition(current_block.position() + self.position_in_block)
        self.block_state = current_block.userState()
        self.document.setTextCursor(current_cursor)
        if current_block.userData():
            block_data = current_block.userData()
        else:
            block_data = BlockUserData()
        if not block_data["style"]:
            block_data["style"] = next(iter(self.document.txt_formats))
        log.debug("Insert: Insert text at %s" % str(current_block.position() + self.position_in_block))
        block_data["strokes"].insert_steno(self.position_in_block, self.steno)
        block_data = update_user_data(block_data, "edittime")
        current_block.setUserData(block_data)
        cursor_format = self.document.txt_formats[block_data["style"]]
        for el in self.steno:
            cursor_format.setForeground(self.document.highlight_colors[el.element])
            # current_cursor.setCharFormat(cursor_format)
            current_cursor.insertText(el.to_text(), cursor_format)
        current_block.setUserState(1)
        self.document.setTextCursor(current_cursor)
        log_dict = {"action": "insert", "block": self.block, "position_in_block": self.position_in_block, "steno": self.steno.to_json()}
        log.info(f"Insert: {log_dict}")
        self.setText("Insert: %s" % self.steno.to_text())
    def undo(self):
        current_cursor = self.document.textCursor()
        current_block = self.document.document().findBlockByNumber(self.block)
        start_pos = current_block.position() + self.position_in_block
        end_pos = current_block.position() + self.position_in_block + len(self.steno.to_text().rstrip("\n"))
        current_cursor.setPosition(start_pos)
        current_cursor.setPosition(end_pos, QTextCursor.KeepAnchor)
        self.document.setTextCursor(current_cursor)
        block_data = current_block.userData()
        res = block_data["strokes"].remove_steno(self.position_in_block, self.position_in_block + len(self.steno.to_text()))
        block_data = update_user_data(block_data, "edittime")
        current_block.setUserData(block_data)
        current_cursor.removeSelectedText()
        log_dict = {"action": "remove", "block": self.block, "position_in_block": self.position_in_block, "end": self.position_in_block + len(self.steno.to_text())}
        log.info(f"Insert (undo): {log_dict}")
        self.document.setTextCursor(current_cursor)
              
class steno_remove(QUndoCommand):
    """Removes text and steno data from editor.

    :param cursor: a ``QTextCursor`` instance
    :param document: a ``QTextDocument`` to act upon
    :param int block: ``blockNumber`` of the ``QTextDocument`` to act upon
    :param int position_in_block: position in the block to act upon
    :param int length: number of characters to remove
    :param steno: holder of removed data
    """
    def __init__(self, cursor, document, block, position_in_block, length, steno = ""):
        super().__init__()
        self.document = document
        self.block = block
        self.position_in_block = position_in_block
        self.length = length
        self.steno = steno
        self.block_state = 1
        self.cursor = cursor
    def redo(self):
        current_cursor = self.cursor
        current_block = self.document.document().findBlockByNumber(self.block)
        start_pos = current_block.position() + self.position_in_block
        current_cursor.setPosition(start_pos)
        block_data = current_block.userData()
        self.block_state = current_block.userState()
        self.steno = block_data["strokes"].remove_steno(self.position_in_block, self.position_in_block + self.length)
        block_data = update_user_data(block_data, "edittime")
        current_cursor.setPosition(start_pos + len(self.steno), QTextCursor.KeepAnchor)
        self.document.setTextCursor(current_cursor)        
        current_block.setUserData(block_data)
        current_cursor.removeSelectedText()
        current_block.setUserState(1)
        self.document.setTextCursor(current_cursor)
        log_dict = {"action": "remove", "block": self.block, "position_in_block": self.position_in_block, "end": self.position_in_block + self.length}
        log.info(f"Remove: {log_dict}")
        self.setText("Remove: %d backspace(s)" % len(self.steno))
    def undo(self):
        current_cursor = self.document.textCursor()
        current_block = self.document.document().findBlockByNumber(self.block)
        start_pos = current_block.position() + self.position_in_block
        end_pos = current_block.position() + self.position_in_block + self.length
        current_cursor.setPosition(start_pos)
        self.document.setTextCursor(current_cursor)
        block_data = current_block.userData()
        res = block_data["strokes"].insert_steno(self.position_in_block, self.steno)
        block_data = update_user_data(block_data, "edittime")
        current_block.setUserData(block_data)
        cursor_format = self.document.txt_formats[block_data["style"]]
        for el in self.steno:
            cursor_format.setForeground(self.document.highlight_colors[el.element])
            # current_cursor.setCharFormat(cursor_format)
            current_cursor.insertText(el.to_text(), cursor_format)        
        log_dict = {"action": "insert", "block": self.block, "position_in_block": self.position_in_block, "steno": self.steno.to_json()}
        log.info(f"Remove (undo): {log_dict}")
        self.document.setTextCursor(current_cursor)

class image_insert(QUndoCommand):
    """Insert image into editor.
    
    :param cursor: a ``QTextCursor`` instance
    :param document: a ``QTextDocument`` to act upon
    :param int block: ``blockNumber`` of the ``QTextDocument`` to act upon
    :param int position_in_block: position in the block to act upon
    :param image_element: an ``image_text`` element to insert
    
    """
    def __init__(self, cursor, document, block, position_in_block, image_element):
        super().__init__()
        self.cursor = cursor
        self.document = document        
        self.block = block
        self.position_in_block = position_in_block
        self.image_element = image_element
        self.block_state = 1
    def redo(self):
        # prep image for qt insert
        asset_dir_path = self.document.file_name / "assets"
        asset_dir_path.mkdir(exist_ok = True)
        asset_dir_name = asset_dir_path / pathlib.Path(self.image_element.path).name
        if not asset_dir_name.exists():
            copyfile(self.image_element.path, asset_dir_name)
        self.image_element.path = asset_dir_name.as_posix()
        # double check and only use path to assets
        imageUri = QUrl(pathlib.Path(self.image_element.path).as_uri())
        image = QImage(QImageReader(self.image_element.path).read())
        self.document.document().addResource(
            QTextDocument.ImageResource,
            imageUri,
            QVariant(image)
        )
        imageFormat = QTextImageFormat()
        imageFormat.setWidth(image.width())
        imageFormat.setHeight(image.height())
        imageFormat.setName(imageUri.toString())
        self.image_element.width = image.width()
        self.image_element.height = image.height()
        current_block = self.cursor.block()
        current_block.userData()["strokes"].insert_steno(self.position_in_block, self.image_element)
        self.block_state = current_block.userState()
        current_cursor = self.cursor
        current_cursor.setPosition(current_block.position() + self.position_in_block)
        log_dict = {"action": "insert", "block": self.block, "position_in_block": self.position_in_block, "steno": self.image_element.to_json()}
        log.info(f"Insert: {log_dict}")
        current_cursor.insertImage(imageFormat)
        current_block.setUserState(1)
        self.setText("Insert: image object")
        self.document.setTextCursor(current_cursor)
    def undo(self):
        current_cursor = self.document.textCursor()
        current_block = self.document.document().findBlockByNumber(self.block)
        current_cursor.setPosition(current_block.position() + self.position_in_block)
        current_cursor.setPosition(current_block.position() + self.position_in_block + 1, QTextCursor.KeepAnchor)
        current_block.userData()["strokes"].remove_steno(self.position_in_block, self.position_in_block + 1)
        current_cursor.removeSelectedText()
        log_dict = {"action": "remove", "block": self.block, "position_in_block": self.position_in_block, "end": self.position_in_block + 1}
        log.info(f"Insert image (undo): {log_dict}")        
        self.document.setTextCursor(current_cursor)

class split_steno_par(QUndoCommand):
    """ Splits a paragraph at position in block, and puts steno properly with new paragraph.

    .. warning::
        Use cursor movements, and not insertText and removeSelectedText
        as otherwise image elements are replaced by the text equivalent.

    :param cursor: a ``QTextCursor`` instance
    :param document: a ``QTextDocument`` to act upon
    :param int block: ``blockNumber`` of the ``QTextDocument`` to act upon
    :param int position_in_block: position in the block to act upon
    :param str space_placement: from Plover config, whether space is before/after output
    :param new_line_stroke: a ``stroke_text`` element containing custom data to insert in the split
    :param bool remove_space: trim space from start of new paragraph after split, default ``True``
    :param bool space_removed: tracks whether a space was removed from start of new paragraph
    """
    def __init__(self, cursor, document, block, position_in_block, space_placement, new_line_stroke, remove_space = True):
        super().__init__()
        self.block = block
        self.position_in_block = position_in_block
        self.document = document
        self.remove_space = remove_space
        self.space_placement = space_placement
        self.new_line_stroke = automatic_text()
        self.new_line_stroke.from_dict(new_line_stroke.to_json())
        self.block_text = ""
        self.block_data = ""
        self.block_state = 1
        self.cursor = cursor
        self.space_removed = False
    def redo(self):
        current_cursor = self.cursor
        current_block = self.document.document().findBlockByNumber(self.block)
        current_cursor.setPosition(current_block.position() + self.position_in_block)
        self.block_data = deepcopy(current_block.userData().return_all())
        self.block_text = current_block.text()
        stroke_data = self.block_data["strokes"]
        # first_part = stroke_data.extract_steno(0, self.position_in_block)
        second_part = current_block.userData()["strokes"].remove_steno(self.position_in_block, len(current_block.userData()["strokes"].to_text()))
        first_data = current_block.userData()
        second_data = BlockUserData()
        first_data = update_user_data(first_data, key = "edittime")
        second_data = update_user_data(second_data, key = "edittime")
        if second_part and self.remove_space and self.space_placement == "Before Output":
            if second_part.starts_with(" "):
                log.debug("Split: Stripping space from beginning of second piece")
                second_part.remove_begin(" ")
                current_cursor.deleteChar()
                self.space_removed = True
        log.debug("Split: Appending new line stroke to end of second piece")
        fake_newline_stroke = self.new_line_stroke
        current_block.userData()["strokes"].append(deepcopy(fake_newline_stroke))
        # this is important on last segment that has audioendtime
        if first_data["audioendtime"]:
            second_data = update_user_data(second_data, key = "audioendtime", value = first_data["audioendtime"])
            first_data = update_user_data(first_data, key = "audioendtime", value = "")
        # check the strokes if there is an audio time associated with stroke
        # use first stroke that has an audio time
        focal_stroke = next((stroke for stroke in second_part if isinstance(stroke, stroke_text) and stroke.audiotime), None) 
        if focal_stroke:
            second_data = update_user_data(second_data, key = "audiostarttime", value = focal_stroke.audiotime)
        # update creationtime
        try:
            second_data = update_user_data(second_data, key = "creationtime", value = second_part.data[0].time)
        except:
            second_data = update_user_data(second_data, key = "creationtime")
        second_data["strokes"] = second_part
        second_data["style"] = first_data["style"]
        # first_data["strokes"] = first_part
        # mimic auto text
        current_cursor.insertText(fake_newline_stroke.to_text().rstrip("\n"))
        current_cursor.insertBlock()
        new_block = current_block.next()
        current_block.setUserData(first_data)
        new_block.setUserData(second_data)
        new_block.setUserState(1)
        self.setText("Split: paragraph %d at %d" % (self.block, self.position_in_block))
        log_dict = {"action": "split", "block": self.block, "position_in_block": self.position_in_block}
        log.info(f"Split: {log_dict}")
        self.block_state = current_block.userState()
        current_block.setUserState(1)
        current_cursor.movePosition(QTextCursor.StartOfBlock)
        self.document.setTextCursor(current_cursor)       
    def undo(self):
        current_cursor = self.document.textCursor()
        current_block = self.document.document().findBlockByNumber(self.block)
        current_cursor.setPosition(current_block.position())
        current_cursor.movePosition(QTextCursor.EndOfBlock)
        current_cursor.deleteChar()
        # removal of the auto text
        for char in self.new_line_stroke.to_text().rstrip("\n"):
            current_cursor.deletePreviousChar()
        # remove the starting space that was removed
        if self.space_removed:
            current_cursor.insertText(" ")
        restore_data = BlockUserData()
        for key, item in self.block_data.items():
            restore_data = update_user_data(restore_data, key = key, value = item)
        current_block.setUserData(restore_data)
        log_dict = {"action": "merge", "block": self.block}
        log.info(f"Split (undo): {log_dict}")
        self.document.setTextCursor(current_cursor)

class merge_steno_par(QUndoCommand):
    """Combine two paragraphs into one in editor.

    .. warning::
        Use cursor movements, and not insertText and removeSelectedText
        as otherwise image elements are replaced by the text equivalent.

    :param cursor: a ``QTextCursor`` instance
    :param document: a ``QTextDocument`` to act upon
    :param int block: ``blockNumber`` of the ``QTextDocument`` to act upon
    :param int position_in_block: position in the block to act upon    
    :param str space_placement: from Plover config, whether space is before/after output
    :param bool add_space: whether to add space between the two paragraphs at merging
    """
    def __init__(self, cursor, document, block, position_in_block, space_placement, add_space = True):
        super().__init__()
        self.block = block
        self.document = document
        self.position_in_block = position_in_block
        self.space_placement = space_placement
        self.add_space = add_space
        self.first_data_dict = {}
        self.second_data_dict = {} 
        self.block_state = 1   
        self.cursor = cursor    
    def redo(self):
        current_cursor = self.cursor
        first_block_num = self.block
        second_block_num = self.block + 1
        if current_cursor.block().blockNumber() == self.block:
            first_block = current_cursor.block()
        else:
            first_block = self.document.document().findBlockByNumber(first_block_num)
        second_block = first_block.next()
        first_data = first_block.userData()
        second_data = second_block.userData()
        self.second_data_dict = deepcopy(second_data.return_all())
        self.first_data_dict = deepcopy(first_data.return_all())
        self.position_in_block = len(first_block.text())
        first_data = update_user_data(first_data, key = "edittime")
        # append second stroke data to first
        second_strokes = second_data["strokes"]
        first_strokes = first_data["strokes"]
        current_cursor.setPosition(first_block.position())
        current_cursor.movePosition(QTextCursor.EndOfBlock)
        if first_data["strokes"].ends_with_element("automatic"):
            for char in first_data["strokes"].data[-1].prefix:
                current_cursor.deletePreviousChar()
        if first_data["strokes"].ends_with("\n"):
            first_data["strokes"].remove_end("\n")
            log.debug("Merge: Removing new line from first paragraph steno")
        if self.space_placement == "Before Output" and self.add_space:
            second_data["strokes"].add_begin(" ")
        elif self.add_space:
            first_data["strokes"].add_end(" ")
        if self.add_space:
            current_cursor.insertText(" ")
        # append the second data to the first
        first_data["strokes"].extend(second_data["strokes"])
        # update the audio end time if it exists, only really needed for last block since other blocks do not have audioendtime
        if first_data["audioendtime"] != second_data["audioendtime"]:
            first_data = update_user_data(first_data, key = "audioendtime", value = second_data["audioendtime"])
        first_block.setUserData(first_data)
        self.block_state = first_block.userState()
        first_block.setUserState(1)
        current_cursor.deleteChar()
        current_cursor.setPosition(first_block.position() + self.position_in_block)
        log_dict = {"action": "merge", "block": self.block}
        log.info(f"Merge: {log_dict}")
        self.document.setTextCursor(current_cursor)
        self.document.refresh_par_style(first_block)
        self.setText("Merge: paragraphs %d & %d" % (first_block_num, second_block_num))
    def undo(self):
        current_cursor = self.document.textCursor()
        first_block_num = self.block
        second_block_num = self.block + 1
        first_block = self.document.document().findBlockByNumber(first_block_num)
        first_data = BlockUserData()
        for key, item in self.first_data_dict.items():
            first_data = update_user_data(first_data, key = key, value = item)
        first_block.setUserData(first_data)
        current_cursor.setPosition(first_block.position())
        if first_data["strokes"].ends_with_element("automatic"):
            cursor_format = self.document.txt_formats[first_data["style"]]
            cursor_format.setForeground(self.document.highlight_colors["automatic"])
            # current_cursor.setCharFormat(cursor_format)
            current_cursor.setPosition(first_block.position() + self.position_in_block - len(first_data["strokes"].data[-1].prefix))
            current_cursor.insertText(first_data["strokes"].data[-1].prefix, cursor_format)
        else:
            current_cursor.setPosition(first_block.position() + self.position_in_block)
        current_cursor.insertText("\n")
        if self.add_space:
            current_cursor.deleteChar()
        second_block = self.document.document().findBlockByNumber(second_block_num)
        second_data = BlockUserData()
        for key, item in self.second_data_dict.items():
            second_data = update_user_data(second_data, key = key, value = item)
        second_block.setUserData(second_data)
        second_block.setUserState(1)
        log_dict = {"action": "split", "block": self.block, "position_in_block": self.position_in_block}
        log.info(f"Merge (undo): {log_dict}")        
        self.document.setTextCursor(current_cursor)
        self.document.refresh_par_style(second_block)

class set_par_style(QUndoCommand):
    """Set paragraph style.

    Character formats have to be applied through the iterator of ``QTextBlock``
    on individual ``QTextFragment`` elements to avoid applying a format on an image,
    over-riding its format, and causing it to revert to an object replacement charater.
    
    :param cursor: a ``QTextCursor`` instance
    :param document: a ``QTextDocument`` to act upon
    :param int block: ``blockNumber`` of the ``QTextDocument`` to act upon
    :param str style: name of style to set
    :param par_formats: ``dict`` containing block-level formats
    :param txt_formats: ``dict`` containing char-level formats

    """
    def __init__(self, cursor, document, block, style, par_formats, txt_formats):
        super().__init__()
        self.cursor = cursor
        self.block = block
        self.style = style
        self.document = document
        self.par_formats = par_formats
        self.txt_formats = txt_formats
        self.old_style = ""
        self.block_state = 1
    def redo(self):
        current_block = self.document.document().findBlockByNumber(self.block)
        block_data = current_block.userData()
        if not block_data:
            block_data = BlockUserData()
        if block_data["style"]:
            self.old_style = block_data["style"]
        # if no style specified, fall back to first style available
        if self.style == "":
            self.style = list(self.par_formats.keys())[0]
        block_data = update_user_data(block_data, "style", self.style)
        current_block.setUserData(block_data)
        self.setText(f"Format: set paragraph {self.block} style to {self.style}")
        self.document.refresh_par_style(current_block)
        self.block_state = current_block.userState()
        current_block.setUserState(1)
        log_dict = {"action": "set_style", "block": self.block, "style": self.style}
        log.info(f"Style: {log_dict}")
    def undo(self):
        if self.old_style:
            current_block = self.document.document().findBlockByNumber(self.block)
            block_data = current_block.userData()
            block_data = update_user_data(block_data, "style", self.old_style)
            current_block.setUserData(block_data)
            self.document.refresh_par_style(current_block)      
            log_dict = {"action": "set_style", "block": self.block, "style": self.old_style}
            log.info(f"Style: {log_dict}")

class set_par_property(QUndoCommand):
    """Set a paragraph's property.

    :param document: a ``QTextDocument`` to act upon
    :param int block: ``blockNumber`` of the ``QTextDocument`` to act upon
    :param str prop: property key
    :param value: value to set for property
    """
    def __init__(self, document, block, prop, value):
        super().__init__()
        self.block = block
        self.prop = prop
        self.new_value = value
        self.old_value = None
        self.document = document
    def redo(self):
        self.setText(f"Property: set paragraph property {self.prop} to {self.new_value}")
        current_block = self.document.document().findBlockByNumber(self.block)
        block_data = current_block.userData()
        self.old_value = block_data[self.prop]
        block_data[self.prop] = self.new_value
        log_dict = {"action": "set_par_prop", "block": self.block, "prop": self.prop, "value": self.new_value}
        log.info(f"Property: {log_dict}")
    def undo(self):
        current_block = self.document.document().findBlockByNumber(self.block)
        block_data = current_block.userData()
        block_data[self.prop] = self.old_value
        log_dict = {"action": "set_par_prop", "block": self.block, "prop": self.prop, "value": self.old_value}
        log.info(f"Property undo: {log_dict}")

class update_style(QUndoCommand):
    """Update entire style.

    :param document: a ``QTextDocument`` to act upon
    :param styles: dict of styles
    :param style_name: name of style to update
    :param new_style_dict: dict of new style parameters

    """
    def __init__(self, document, styles, style_name, new_style_dict):
        super().__init__()
        self.document = document
        self.styles = styles
        self.style_name = style_name
        self.new_style_dict = deepcopy(new_style_dict)
        self.old_style_dict = {}
    def redo(self):
        if self.style_name in self.styles:
            self.old_style_dict = deepcopy(self.styles[self.style_name])
        self.styles[self.style_name] = self.new_style_dict
        log_dict = {"action": "edit_style", "style_dict": self.new_style_dict}
        log.info(f"Style: {log_dict}")
        self.setText(f"Style: update style attributes for style {self.style_name}")
        self.document.gen_style_formats()
    def undo(self):
        self.styles[self.style_name] = self.old_style_dict
        log_dict = {"action": "edit_style", "style_dict": self.old_style_dict}
        log.info(f"Style undo: {log_dict}")
        self.document.gen_style_formats()

class update_config_value(QUndoCommand):
    """Update transcript config value.

    :param str key: config key
    :param str value: config value
    :param dict config: transcript config

    """
    def __init__(self, key, value, config):
        super().__init__()
        self.config_key = key
        self.old_value = None
        self.new_value = value
        self.config = config
    def redo(self):
        self.old_value = deepcopy(self.config[self.config_key])
        self.config[self.config_key] = self.new_value
        log_dict = {"action": "config", "key": self.config_key, "value": self.new_value}
        log.info(f"Config: {log_dict}")
        self.setText("Config: updated value.")
    def undo(self):
        self.config[self.config_key] = self.old_value
        log_dict = {"action": "config", "key": self.config_key, "value": self.old_value}
        log.info(f"Config (undo): {log_dict}")

class update_field(QUndoCommand):
    """Update transcript fields with new values.

    When fields change values, the underlying ``text_field``
    will be updated when called again, but the text
    in the ``QTextEdit`` will not. This command updates 
    the ``text_field`` element, then removes the old field value 
    and inserts the new one into text. 

    The original ``dict`` that has to be updated is outside 
    the command and has to be passed in by reference. 
    But a copy each of the old and the new dicts are needed
    to perform redo/undos properly.

    :param cursor: a ``QTextCursor`` instance
    :param document: a ``QTextDocument`` to act upon
    :param int block: ``blockNumber`` of the ``QTextDocument`` to act upon
    :param int position_in_block: position in the block to act upon
    :param dict old_dict: dict with existing field values
    :param new_dict: dict with new field values
    """
    def __init__(self, cursor, document, block, position_in_block, old_dict, new_dict):
        super().__init__()
        self.document = document
        self.block = block
        self.position_in_block = position_in_block
        # first is the reference to dict to be updated
        self.user_field_dict = deepcopy(old_dict)
        self.new_dict = deepcopy(new_dict)
        # second one is the copy to be kept for undos
        self.store_dict = deepcopy(old_dict)
        self.cursor = cursor
    def redo(self):
        current_cursor = self.cursor
        current_block = self.document.document().findBlockByNumber(self.block)
        block = self.document.document().begin()
        self.document.user_field_dict = self.new_dict
        self.document.set_config_value("user_field_dict", self.new_dict)
        for i in range(self.document.document().blockCount()):     
            block_strokes = block.userData()["strokes"]
            if any([el.element == "field" for el in block_strokes]):
                block.setUserState(1)
                for ind, el in enumerate(block_strokes):
                    # print(ind)
                    if el.element == "field":
                        start_pos, end_pos = block_strokes.element_pos(ind)
                        # print(block.position() + start_pos)
                        current_cursor.setPosition(block.position() + start_pos)
                        current_cursor.setPosition(block.position() + end_pos, QTextCursor.KeepAnchor)
                        current_cursor.removeSelectedText()
                        el.user_dict = self.new_dict
                        cursor_format = self.document.txt_formats[block.userData()["style"]]
                        cursor_format.setForeground(self.document.highlight_colors["index"])
                        current_cursor.insertText(el.to_text(), cursor_format)
            if block == self.document.document().lastBlock():
                break
            block = block.next()
        current_cursor.setPosition(current_block.position() + self.position_in_block)
        self.setText("Fields: update fields")
        log_dict = {"action": "field", "field": self.new_dict}
        log.info(f"Field: {log_dict}")
    def undo(self):
        current_cursor = self.document.textCursor()
        current_block = self.document.document().findBlockByNumber(self.block)
        block = self.document.document().begin()
        self.document.user_field_dict = self.store_dict
        self.document.set_config_value("user_field_dict", self.store_dict)
        for i in range(self.document.document().blockCount()):   
            block_strokes = block.userData()["strokes"]
            for ind, el in enumerate(block_strokes):
                # print(ind)
                if el.element == "field":
                    start_pos, end_pos = block_strokes.element_pos(ind)
                    # print(block.position() + start_pos)
                    current_cursor.setPosition(block.position() + start_pos)
                    current_cursor.setPosition(block.position() + end_pos, QTextCursor.KeepAnchor)
                    current_cursor.removeSelectedText()
                    el.user_dict = self.user_field_dict
                    cursor_format = self.document.txt_formats[block.userData()["style"]]
                    cursor_format.setForeground(self.document.highlight_colors["index"])
                    # current_cursor.setCharFormat(cursor_format)  
                    current_cursor.insertText(el.to_text(), cursor_format)
            if block == self.document.document().lastBlock():
                break
            block = block.next()
        current_cursor.setPosition(current_block.position() + self.position_in_block)
        log_dict = {"action": "field", "field": self.store_dict}
        log.info(f"Field: {log_dict}")        

class update_entries(QUndoCommand):
    """Update index entries with new values.

    :param cursor: a ``QTextCursor`` instance
    :param document: a ``QTextDocument`` to act upon
    :param int block: ``blockNumber`` of the ``QTextDocument`` to act upon
    :param int position_in_block: position in the block to act upon
    :param dict old_dict: dict with existing index data
    :param new_dict: dict with new index data    
    """
    def __init__(self, cursor, document, block, position_in_block, old_dict, new_dict):
        super().__init__()
        self.document = document
        self.block = block
        self.position_in_block = position
        self.new_dict = deepcopy(new_dict)
        # second one is the copy to be kept for undos
        self.store_dict = deepcopy(old_dict)
        self.cursor = cursor
    def redo(self):
        current_cursor = self.cursor
        current_block = self.document.document().findBlockByNumber(self.block)
        block = self.document.document().begin()
        for i in range(self.document.document().blockCount()):    
            block_strokes = block.userData()["strokes"]
            if any([el.element == "index" for el in block_strokes]):
                block.setUserState(1)
                for ind, el in enumerate(block_strokes):
                    # print(ind)
                    if el.element == "index":
                        start_pos, end_pos = block_strokes.element_pos(ind)
                        current_cursor.setPosition(block.position() + start_pos)
                        current_cursor.setPosition(block.position() + end_pos, QTextCursor.KeepAnchor)
                        current_cursor.removeSelectedText()
                        el.prefix = self.new_dict[el.indexname]["prefix"]
                        el.hidden = self.new_dict[el.indexname]["hidden"]
                        el.description = self.new_dict[el.indexname]["entries"][el.data]
                        cursor_format = self.document.txt_formats[block.userData()["style"]]
                        cursor_format.setForeground(self.document.highlight_colors["index"])
                        # current_cursor.setCharFormat(cursor_format)                        
                        current_cursor.insertText(el.to_text(), cursor_format)
            if block == self.document.document().lastBlock():
                break
            block = block.next()
        current_cursor.setPosition(current_block.position() + self.position_in_block)
        self.setText("Indices: update indices.")
        log_dict = {"action": "index", "index": self.new_dict}
        log.info(f"Index: {log_dict}")
    def undo(self):
        current_cursor = self.document.textCursor()
        current_block = self.document.document().findBlockByNumber(self.block)
        block = self.document.document().begin()
        for i in range(self.document.document().blockCount()):    
            block_strokes = block.userData()["strokes"]
            for ind, el in enumerate(block_strokes):
                # print(ind)
                if el.element == "index":
                    start_pos, end_pos = block_strokes.element_pos(ind)
                    current_cursor.setPosition(block.position() + start_pos)
                    current_cursor.setPosition(block.position() + end_pos, QTextCursor.KeepAnchor)
                    current_cursor.removeSelectedText()
                    el.prefix = self.store_dict[el.indexname]["prefix"]
                    el.hidden = self.store_dict[el.indexname]["hidden"]
                    el.description = self.store_dict[el.indexname]["entries"][el.data]
                    cursor_format = self.document.txt_formats[block.userData()["style"]]
                    cursor_format.setForeground(self.document.highlight_colors["index"])
                    current_cursor.insertText(el.to_text(), cursor_format)
            if block == self.document.document().lastBlock():
                break
            block = block.next()
        current_cursor.setPosition(current_block.position() + self.position_in_block)
        log_dict = {"action": "index", "index": self.store_dict}
        log.info(f"Index: {log_dict}")  


from PyQt5 import QtCore, QtGui, QtWidgets
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QDialog, QMessageBox
from PyQt5.QtGui import QStandardItemModel, QKeySequence
from plover_cat.shortcut_dialog_ui import Ui_shortcutDialog

class shortcutDialogWindow(QDialog, Ui_shortcutDialog):
    """Set shortcuts for menu items in editor through the modal dialog.

    :param shortcut_dict: existing dict of shortcuts, {action: "str of keys"}
    :type shortcut_dict: dict
    :param menu_names: visible text of QAction objects in menus
    :type menu_names: list
    :param action_names: object names of QAction objects in menus
    :type action_names: list
    """
    def __init__(self, shortcut_dict, menu_names, action_names):
        super().__init__()
        self.setupUi(self)
        self.shortcut_dict = shortcut_dict
        self.menu_names = menu_names
        self.action_names = action_names
        self.text_name.clear()
        for menu, action in zip(self.menu_names, self.action_names):
            self.text_name.addItem(menu, action)
        if self.text_name.currentText():
            self.display_shortcut(self.text_name.currentIndex())
        self.validate.clicked.connect(self.check_save_shortcut)
        self.text_name.activated.connect(self.display_shortcut)          
    def display_shortcut(self, index):
        """Display existing shortcut, if set, of selected menu item"""
        self.shortcut.clear()
        self.text_name.setCurrentIndex(index)
        action = self.text_name.currentData()
        if action in self.shortcut_dict:
            keys = QKeySequence(self.shortcut_dict[action])
            self.shortcut.setKeySequence(keys)
    def check_save_shortcut(self):
        """Check if shortcut displayed is valid"""
        action = self.text_name.currentData()
        if not action:
            return
        keys = self.shortcut.keySequence().toString()
        if keys in self.shortcut_dict.values():
            QMessageBox.warning(self, "Plover2CAT", "This shortcut is already in use. Choose another one.")
            return
        reserved_keys = [f"Ctrl+{i}" for i in range(0,10)] + [f"Ctrl+Shift+{i}" for i in range(0,10)] 
        if keys in reserved_keys:
            QMessageBox.warning(self, "Plover2CAT", "This shortcut is a reserved shortcut. Choose another one.")
            return
        self.shortcut_dict[action] = keys